# suggested_card.py

"""Suggested card for Card Player."""

from bridgeobjects import Card, SEATS
from bfgdealer import Board

from .player import Player
from .opening_lead_card import opening_lead_card
from .opening_lead_suit import opening_lead_suit
from .first_seat_defender import FirstSeatDefender
from .first_seat_declarer_suit import FirstSeatDeclarerSuit
from .first_seat_declarer_nt import FirstSeatDeclarerNT
from .second_seat_defender import SecondSeatDefender
from .second_seat_declarer import SecondSeatDeclarer
from .third_seat_defender import ThirdSeatDefender
from .third_seat_declarer import ThirdSeatDeclarer
from .fourth_seat_defender import FourthSeatDefender
from .fourth_seat_declarer import FourthSeatDeclarer

import bfgcardplay.src.global_variables as global_vars

MODULE_COLOUR = 'green'


def next_card(board: Board) -> Card | None:
    """Return the suggested card from the current status of the board."""
    unplayed_cards_in_board = _unplayed_cards(board)

    if unplayed_cards_in_board == 52:
        return _opening_lead(board)

    if unplayed_cards_in_board == 0:
        return None

    return _next_card(board)


def _next_card(board: Board) -> Card:
    player = Player(board)
    trick_cards = len(board.tricks[-1].cards)
    if trick_cards in [0, 4]:
        # TODO remove print statement
        # if not board.status == Board.STATUS['Test']:
        #     print('~'*50)
        return first_seat_card(player)
    if trick_cards == 1:
        return second_seat_card(player)
    if trick_cards == 2:
        return third_seat_card(player)
    if trick_cards == 3:
        return fourth_seat_card(player)
    raise ValueError(f'No card to return for {board}')


def _opening_lead(board: Board) -> Card:
    """Return the proposed opening lead for the board."""
    global_vars.initialize()
    if not board.contract.declarer:
        return None
    # TODO remove print statement
    # print('#'*50)
    opening_suit = opening_lead_suit(board)
    leader = board.contract.leader
    leaders_cards = board.hands[leader].cards
    cards = [card for card in leaders_cards if card.suit == opening_suit]
    opening_card = opening_lead_card(cards, board.contract)
    return opening_card


def first_seat_card(player: Player) -> Card:
    """Return the card for the first seat."""
    if player.is_defender:
        selected_card = FirstSeatDefender(player).selected_card()
    else:
        if player.trump_suit:
            selected_card = FirstSeatDeclarerSuit(player).selected_card()
        else:
            selected_card = FirstSeatDeclarerNT(player).selected_card()
    return selected_card


def second_seat_card(player: Player) -> Card:
    """Return the card for the second seat."""
    if player.is_defender:
        selected_card = SecondSeatDefender(player).selected_card()
    else:
        selected_card = SecondSeatDeclarer(player).selected_card()
    return selected_card


def third_seat_card(player: Player) -> Card:
    """Return the card for the third seat."""
    if player.is_defender:
        selected_card = ThirdSeatDefender(player).selected_card()
    else:
        selected_card = ThirdSeatDeclarer(player).selected_card()
    return selected_card


def fourth_seat_card(player: Player) -> Card:
    """Return the card for the fourth seat."""
    if player.is_defender:
        selected_card = FourthSeatDefender(player).selected_card()
    else:
        selected_card = FourthSeatDeclarer(player).selected_card()
    return selected_card


def _unplayed_cards(board: Board):
    unplayed_cards = 0
    for seat in SEATS:
        unplayed_cards += len(board.hands[seat].unplayed_cards)
    return unplayed_cards



""" Generate and play random hands until it breaks."""
import sys
from termcolor import cprint
import logging
from io import StringIO
import gi
gi.require_version('Notify', '0.7')
from gi.repository import Notify

from bridgeobjects import Card, Trick, SEATS
from bfgdealer import Dealer, Board

from bfgcardplay.src.suggested_card import next_card
# from application.app.folder.file import func_name

MODULE_COLOUR = 'blue'
MAX_HANDS = 10**6


def main():
    try:
        generate_hands()
    except:
        log_stream = StringIO()
        logging.basicConfig(stream=log_stream, level=logging.ERROR)
        logging.error("Exception occurred", exc_info=True)
        print('x'*100)
        print(log_stream.getvalue())
        print('x'*100)

        Notify.init("Card play blitz")
        Notify.Notification.new('Error occurred').show()

def generate_hands():
    # logging.disable(logging.CRITICAL)
    count = 0
    while count < MAX_HANDS:
        board = Dealer().deal_random_board()
        board.auction = board.get_auction()
        board.contract = board.get_contract()
        if not board.contract.name:
            continue

        display_board(board, count)

        for _ in range(52):
            card = next_card(board)
            update_board_after_cardplay(board, card)
        # halt()

        count += 1


def display_board(board, count):
    if count % 10 == 0:
        print('.', end='', flush=True)
    if count and count % 1000 == 0:
        print(count)
        print('.', end='', flush=True)
        Notify.init("Card play blitz")
        Notify.Notification.new(f'{count} boards processed').show()
    # os.system('clear')
    # print(f"{count + 1} {'-'*50}")
    # cprint('\n'.join(board.board_to_pbn()), MODULE_COLOUR)


def update_board_after_cardplay(board: Board, card: Card) -> Board:
    trick = board.tricks[-1]
    current_player = get_current_player(trick)
    if card and card in board.hands[current_player].unplayed_cards:
        add_card_to_trick(board, trick, card)
        board.hands[current_player].unplayed_cards.remove(card)
    return board


def add_card_to_trick(board, trick, card):
    if card in trick.cards:
        return
    trick.cards.append(card)
    if len(trick.cards) == 4:
        complete_trick(board, trick)


def complete_trick(board: Board, trick: Trick) -> str:
    """complete the trick update the board and return trick winner."""
    trick.complete(board.contract.denomination)
    winner = trick.winner
    trick = Trick(leader=winner)
    board.tricks.append(trick)
    return winner


def get_current_player(trick: Trick) -> str:
    """Return the current player from the trick."""
    if len(trick.cards) == 4:
        return trick.winner
    leader_index = SEATS.index(trick.leader)
    current_player = (leader_index + len(trick.cards)) % 4
    return SEATS[current_player]


def halt():
    result = input('-->')
    if result in ['Q', 'q']:
        sys.exit()


if __name__ == '__main__':
    main()



"""Fourth seat card play."""
from termcolor import cprint

import inspect
from ..logger import log

from bridgeobjects import Card, Denomination
from bfgdealer import Trick
from .player import Player
from .utilities import trick_card_values

MODULE_COLOUR = 'blue'


class FourthSeat():
    def __init__(self, player: Player):
        self.player = player
        self.trick = player.board.tricks[-1]
        self.cards = player.cards_for_trick_suit(self.trick)

    def _winning_card(self, trick: Trick) -> Card | None:
        """Return the card if can win trick."""
        player = self.player
        cards = self.cards

        values = trick_card_values(trick, player.trump_suit)
        if cards:
            for card in cards[::-1]:
                card_value = card.value
                if card.suit == player.trump_suit:
                    card_value += 13
                if card_value > values[0] and card_value > values[2]:
                    log(inspect.stack(), f'{card}')
                    return card

        # No cards in trick suit, look for trump winner
        elif player.trump_cards:
            for card in player.trump_cards[::-1]:
                if card.value + 13 > values[2]:
                    log(inspect.stack(), f'{card}')
                    return card
        return None

    def _second_player_winning_trick(self, cards: list[Card],
                                     trick: Trick, trumps: Denomination) -> bool:
        """Return True if the second player is winning the trick."""
        values = trick_card_values(trick, trumps)
        if values[1] > values[0] and values[1] > values[2]:
            return True
        return False

    def can_win_trick(self, player, card) -> bool:
        """Return True if card can win trick."""
        trick = player.board.tricks[-1]
        trumps = player.trump_suit
        values = trick_card_values(trick, trumps)
        if values[1] > values[0] and values[1] > values[2]:
            return False
        if card.value > values[0] and card.value > values[2]:
            return True
        return False



# first_seat_declarer.py
""" First seat card play for declarer."""

from termcolor import cprint

import inspect
from ..logger import log

from bridgeobjects import SUITS, Card, Suit

import bfgcardplay.src.global_variables as global_vars
from .utilities import get_list_of_best_scores
from .player import Player
from .first_seat import FirstSeat
from .declarer_play import card_combinations

ALL_WINNERS = 5

MODULE_COLOUR = 'cyan'


def return_and_print(obj: object, text: str, can_print: bool) -> object:
    """Print text and return object."""
    if can_print:
        cprint(f'{obj} {text}', MODULE_COLOUR)
    return obj


class FirstSeatDeclarerSuit(FirstSeat):
    def __init__(self, player: Player):
        super().__init__(player)

    def selected_card(self) -> Card:
        """Return the card for the first seat."""
        player = self.player

        if len(player.hand.unplayed_cards) == 1:
            return log(inspect.stack(), player.hand.unplayed_cards[0])

        card = self._playout_winners()
        if card:
            return card

        card = self._manager_force_play_card()
        if card:
            return card

        card = self._card_from_suit_to_develop()
        if card:
            return card

        suit = self._select_suit()
        if not player.suit_cards[suit.name]:
            raise ValueError(f'No cards for {suit}')
        card = self._select_lead_card(suit)
        return card

    def _playout_winners(self):
        # +1 because the next trick has already been created
        player = self.player
        if player.winner_count >= 13 - len(player.board.tricks) + 1:
            card = self._play_all_winners()
            if card:
                return card
        return None

    def _manager_force_play_card(self) -> Card:
        player = self.player
        manager = global_vars.manager
        self.update_manager()

        if manager.card_to_play[player.seat]:
            card = manager.card_to_play[player.seat][0]
            manager.card_to_play[player.seat].pop(0)
            if manager.suit_to_develop(player.seat) and manager.card_to_play[player.seat]:
                manager.card_to_play[player.seat] = player.suit_cards[manager.suit_to_develop][-1]
            return log(inspect.stack(), card)
        return None

    def _card_from_suit_to_develop(self) -> Card:
        player = self.player
        manager = global_vars.manager
        suit_to_develop = manager.suit_to_develop(player.seat)
        if suit_to_develop:
            if player.unplayed_cards[suit_to_develop.name]:
                if manager.suit_strategy(player.seat)[suit_to_develop.name] == 'ruff_in_short_hand':
                    if player.threats[suit_to_develop.name]:
                        if player.is_declarer:
                            my_cards = player.unplayed_cards[suit_to_develop.name]
                            dummys_cards = player.dummys_unplayed_cards[suit_to_develop.name]
                            if dummys_cards and dummys_cards[0].value > my_cards[0].value:
                                return log(inspect.stack(), my_cards[-1])
                        card = player.unplayed_cards[suit_to_develop.name][0]
                        return log(inspect.stack(), card)
        return None

    def _select_suit(self) -> Suit:
        """Return the trick lead suit for the declarer in  a suit contract."""
        player = self.player
        manager = global_vars.manager

        # Draw trumps?
        can_draw_trumps = self._can_draw_trumps()
        if player.opponents_trumps and can_draw_trumps:
            return log(inspect.stack(), player.trump_suit)

        # Suit to develop
        suit = self._define_suit_to_develop()
        if suit:
            return suit
        else:
            self._define_suit_with_winners()

        # Play master
        suit = self._select_suit_with_master()
        if suit:
            return suit

        # Long suits
        suit = self._lead_towards_tenace()
        if suit:
            return suit

        # Play suit to develop
        suit = manager.suit_to_develop(player.seat)
        if suit and player.unplayed_cards[suit.name]:
            return log(inspect.stack(), suit)
        self._update_manager_with_threats()

        # Suit to develop
        suit = self._get_suit_to_develop()
        if suit:
            return suit

        # Lead to tenace
        suit = self._lead_to_tenace()
        if suit:
            return suit

        # Play touching honours
        suit = self._play_touching_honours()
        if suit:
            return suit

        # Look for void in partner's hand
        suit = self._void_in_partners_hand()
        if suit:
            return suit

        # Select best suit
        score_reasons = self._get_score_reasons()
        best_suit = self._best_suit(score_reasons)
        return log(inspect.stack(), best_suit)

    def _get_score_reasons(self) -> dict[str, int]:
        score_reasons = {
            'trumps': self._deprecate_trumps(),
            'shortage': self._shortage_score(self.player.controls),
            'void': self._deprecate_suits(),
            'all_winners': self._all_winners_score(),
            'frozen': self._frozen_suits(),
            'long': self._long_suits(),
        }
        return score_reasons

    def _define_suit_to_develop(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        suit_to_develop = manager.suit_to_develop(player.seat)
        if suit_to_develop and player.unplayed_cards[suit_to_develop]:
            return log(inspect.stack(), suit_to_develop)
        return None

    def _define_suit_with_winners(self) -> Suit:
        # Holds all winners
        player = self.player
        manager = global_vars.manager
        for suit_name, suit in SUITS.items():
            if suit != player.trump_suit:
                if player.holds_all_winners_in_suit(suit):
                    manager.set_suit_to_develop(player.seat, suit)
        return None

    def _select_suit_with_master(self) -> Suit:
        player = self.player
        if not player.opponents_trumps:
            total_cards = 0
            for suit in SUITS:
                holding = player.unplayed_cards[suit]
                total_cards += len(holding)
            if len(player.unplayed_cards[player.trump_suit]) > total_cards - 2:
                return log(inspect.stack(), player.trump_suit)

            for suit in SUITS:
                if suit != player.trump_suit.name and player.unplayed_cards[suit]:
                    if player.is_winner_declarer(player.unplayed_cards[suit][0]):
                        return log(inspect.stack(), Suit(suit))
        return None

    def _lead_towards_tenace(self) -> Suit:
        player = self.player
        long_suits = player.partnership_long_suits()
        for suit in long_suits:
            if not self._frozen_suit(suit) and player.can_lead_toward_tenace(suit):
                return log(inspect.stack(), Suit(suit))
        return None

    def _get_total_winners(self) -> int:
        winners = self._all_winners_score()
        total_winners = 0
        for suit_count in winners:
            total_winners += suit_count[1]
        return total_winners

    def _get_suit_to_develop(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        total_winners = self._get_total_winners()
        if total_winners <= 13 - player.trick_number:
            suit = manager.suit_to_develop(player.seat)
            if suit:
                suit_name = suit.name
                if suit_name and player.suit_cards[suit_name]:
                    return log(inspect.stack(), manager.suit_to_develop(player.seat))
                if not player.suit_cards[suit_name]:
                    manager.set_suit_to_develop(player.seat, None)
        return None

    def _lead_to_tenace(self) -> Suit:
        player = self.player
        if not player.opponents_trumps:
            long_suits = player.partnership_long_suits()
            for suit in long_suits:
                if not self._frozen_suit(suit) and player.can_lead_toward_tenace(suit):
                    return log(inspect.stack(), Suit(suit))
        return None

    def _void_in_partners_hand(self) -> Suit:
        player = self.player
        if player.partners_unplayed_cards[player.trump_suit.name]:
            for suit in SUITS:
                if player.unplayed_cards[suit] and not player.partners_unplayed_cards[suit]:
                    return log(inspect.stack(), Suit(suit))
        return None

    def _play_touching_honours(self) -> Suit:
        player = self.player
        for suit in SUITS:
            can_lead_to_dummy = (player.touching_honours_in_hand(player.dummys_hand, suit) and
                                 player.unplayed_cards[suit])
            if player.touching_honours_in_hand(player.hand, suit) or can_lead_to_dummy:
                return log(inspect.stack(), Suit(suit))
        return None

    def _update_manager_with_threats(self):
        player = self.player
        manager = global_vars.manager
        for suit, threat_cards in player.threats.items():
            ace = Card('A', suit)
            king = Card('K', suit)
            queen = Card('Q', suit)
            if len(threat_cards) == 1:
                threat = threat_cards[0]
                if threat == ace and king in player.unplayed_cards[suit]:
                    manager.set_suit_to_develop(player.seat, Suit(suit))
                    manager.set_suit_strategy(player.seat, suit, 'missing_ace')
                elif threat == king and ace in player.unplayed_cards[suit]:
                    manager.set_suit_to_develop(player.seat, Suit(suit))
                    manager.set_suit_strategy(player.seat, suit, 'missing_king')
                elif threat == queen and king in player.unplayed_cards[suit]:
                    manager.set_suit_to_develop(player.seat, Suit(suit))
                    manager.set_suit_strategy(player.seat, suit, 'missing_queen')

    def _select_lead_card(self, suit: Suit) -> Card:
        """Return the selected lead card for declarer."""
        player = self.player
        my_cards = player.unplayed_cards[suit.name]
        partners_cards = player.partners_unplayed_cards[suit.name]
        partner_long = len(partners_cards) > len(my_cards)

        # Lead low to partner's winner
        if not partner_long and partners_cards and player.is_winner_declarer(partners_cards[0]):
            return log(inspect.stack(), my_cards[-1])

        # if my_cards and not partner_long:
        #     for card in my_cards[len(partners_cards)-1::-1]:
        #         if player.is_winner_declarer(card):
        #             return log(inspect.stack(), card)

        if player.is_winner_declarer(my_cards[0]) and not player.partners_unplayed_cards[suit.name]:
            return log(inspect.stack(), my_cards[0])

        # Top of touching honours
        for index, card in enumerate(my_cards[:-1]):
            if card.is_honour and card.value == my_cards[index+1].value + 1:
                return log(inspect.stack(), card)

        # Top of doubleton
        if len(my_cards) == 2:
            return log(inspect.stack(), my_cards[0])

        # Card combinations
        card = card_combinations(player, suit.name)
        if card:
            return log(inspect.stack(), card)

        # Winners
        for card in my_cards[::-1]:
            if player.is_winner_declarer(card):
                return log(inspect.stack(), card)

        # Return bottom card
        return log(inspect.stack(), my_cards[-1])

    def _get_winning_suit(self) -> Suit:
        """Return the selected winning suit."""
        player = self.player
        winning_suits = self._get_winning_suit_list()

        long_suits = get_list_of_best_scores(winning_suits)
        if long_suits and player.unplayed_cards[long_suits[0]]:
            return log(inspect.stack(), Suit(long_suits[0]))
        return None

    def _get_winning_suit_list(self) -> list[Suit]:
        """Return a list of winning suits."""
        player = self.player
        winning_suits = {}
        for suit_name, suit in SUITS.items():
            our_cards = player.our_unplayed_cards[suit.name]
            if suit != player.trump_suit and our_cards:
                my_cards = player.unplayed_cards[suit_name]
                partners_cards = player.partners_unplayed_cards[suit_name]

                last_card = max(len(my_cards), len(partners_cards)) - 1
                opponents_cards = player.opponents_unplayed_cards[suit.name]
                for index, card in enumerate(our_cards[:last_card]):
                    if index < len(opponents_cards):
                        if card.value < opponents_cards[index].value:
                            break
                else:
                    winning_suits[suit_name] = last_card
        return winning_suits

    def _can_draw_trumps(self) -> tuple[bool, bool]:
        """Return True if declarer should draw trumps, and/or True to ruff in shorthand."""
        show_return = 0
        player = self.player
        if not player.opponents_unplayed_cards[player.trump_suit.name]:
            return return_and_print(False, 'can_draw_trumps z', show_return)
        if not player.trump_cards:
            return return_and_print(False, 'can_draw_trumps a', show_return)

        if self._ruff_in_short_hand():
            return return_and_print(False, 'can_draw_trumps b', show_return)

        # declarers_trumps = Cards(player.declarers_hand.cards).by_suit[player.trump_suit]
        # dummys_trumps = Cards(player.dummys_hand.cards).by_suit[player.trump_suit]
        # if len(declarers_trumps) + len(dummys_trumps) <= 7 and max(len(declarers_trumps), len(dummys_trumps)) == 4:
        #     return return_and_print(False, 'can_draw_trumps c', show_return)

        # declarers_trumps = player.get_unplayed_cards_by_suit(player.trump_suit, player.declarer)
        # dummys_trumps = player.get_unplayed_cards_by_suit(player.trump_suit, player.dummy)
        our_trumps = player.our_unplayed_cards[player.trump_suit]
        if len(our_trumps) >= 9:  # TODO too simplistic - determine losers first
            return return_and_print(True, 'can_draw_trumps d', show_return)

        # short_hand = player.hands[player.short_suit_seat]
        # cprint(f'{short_hand}', MODULE_COLOUR)
        # if short_hand.shape[3] <= 2:
        #     # Are there sufficient trumps in short hand?
        #     opponents_trumps = player.opponents_unplayed_cards[player.trump_suit.name]
        #     if len(short_hand.cards_by_suit[player.trump_suit.name]) > len(opponents_trumps):
        #         return return_and_print(True, 'can_draw_trumps e', show_return)
        #     # can declarer produce a void in time?
        #     suit = short_hand.shortest_suit
        #     our_cards = player.our_unplayed_cards[suit]

        #     ace = Card('A', suit.name)
        #     king = Card('K', suit.name)
        #     if ((ace in player.total_unplayed_cards[suit.name] and ace not in our_cards) and
        #             (king in player.total_unplayed_cards[suit.name] and king not in our_cards)):
        #         return return_and_print(False, 'can_draw_trumps f', show_return)
        return return_and_print(True, 'can_draw_trumps g', show_return)

    def _play_all_winners(self) -> Card:
        """Return a card to play when all are winners"""
        player = self.player

        # Draw trumps
        if player.trump_suit:
            suit_name = player.trump_suit.name
            if player.unplayed_cards[suit_name] and player.opponents_unplayed_cards[suit_name]:
                return self._select_card_from_right_hand(suit_name)
        for suit_name, suit in SUITS.items():
            if player.trump_suit and suit == player.trump_suit:
                continue
            if player.unplayed_cards[suit_name]:
                if player.holds_all_winners_in_suit(suit):
                    return self._select_card_from_right_hand(suit_name)
        return None

    def _select_card_from_right_hand(self, suit_name: str) -> Card:
        """Return the appropriate card depending on suit length."""
        player = self.player
        my_cards = player.unplayed_cards[suit_name]
        partners_cards = player.partners_unplayed_cards[suit_name]
        if (partners_cards and
                len(my_cards) < len(partners_cards) and
                my_cards[-1].value < partners_cards[0].value):
            return log(inspect.stack(), player.unplayed_cards[suit_name][0])
        else:
            for card in my_cards[::-1]:
                if player.is_winner_declarer(card):
                    return log(inspect.stack(), card)
        return log(inspect.stack(), player.unplayed_cards[suit_name][-1])

    def _ruff_in_short_hand(self):
        """Update manager if can ruff in short hand and return True."""
        player = self.player
        manager = global_vars.manager

        (long_trump_hand, short_trump_hand) = player.get_long_short_trump_hands()
        entries = player.get_entries(long_trump_hand)
        total_entries = 0
        for suit in SUITS:
            if suit != player.trump_suit.name:
                total_entries += len(entries[suit])
        if not total_entries:
            return False

        if not player.partners_unplayed_cards[player.trump_suit.name]:
            return False

        for suit in SUITS:
            (long_hand, short_hand) = player.get_long_short_hands(Suit(suit))
            long_cards = long_hand.cards_by_suit[suit]
            short_cards = short_hand.cards_by_suit[suit]
            if (short_cards and
                    len(short_cards) <= 2 and
                    suit != player.trump_suit.name and
                    long_hand == long_trump_hand):
                if ((Card('A', suit) in long_cards and
                        not short_cards[0].is_honour) or
                        Card('A', suit) in short_cards):
                    manager.set_suit_strategy(player.seat, suit, 'ruff_in_short_hand')
                    manager.set_suit_to_develop(player.seat, Suit(suit))
                    if not player.threats[suit]:
                        return False
                    # if not player.is_winner_declarer(player.unplayed_cards[suit][-1]):
                    #     cprint(f'{suit} {player.unplayed_cards[suit]}', MODULE_COLOUR)
                    return True
        return False

    def _long_suits(self) -> list[tuple[str, int]]:
        """Assign score to suits based on length (long)."""
        player = self.player
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        length_score = self.LENGTH_SCORE

        for suit in SUITS:
            number_cards = len(player.our_cards[suit]) - len(player.opponents_cards[suit])
            if number_cards > 0:
                if player.trump_suit:
                    if suit == player.trump_suit.name:
                        length_score = 0
                suit_scores[suit] = number_cards * length_score
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _all_winners_score(self) -> list[tuple[str, int]]:
        """Assign score to a suit if all the cards are winners."""
        player = self.player
        suit_scores = {suit_name: 0 for suit_name, suit in SUITS.items()}
        for suit_name, suit in SUITS.items():
            if suit != player.trump_suit:
                if player.holds_all_winners_in_suit(suit):
                    suit_scores[suit_name] = ALL_WINNERS
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _deprecate_trumps(self) -> list[tuple[str, int]]:
        """Assign score to a suit if opponents have no trumps."""
        suit_scores = {suit_name: 0 for suit_name, suit in SUITS.items()}
        suit_scores[self.player.trump_suit.name] = -1 * self.TRUMP_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _suit_strength(self) -> list[tuple[str, int]]:
        """Assign score to a suit by HCP."""
        suit_scores = {suit_name: 0 for suit_name, suit in SUITS.items()}
        for suit_name in SUITS:
            cards = self.player.our_cards[suit_name]
            for card in cards:
                if card.value >= 10:
                    suit_scores[suit_name] += card.value - 9
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _shortage_score(self, controls: dict[str, int]) -> list[tuple[str, int]]:
        """Assign score to a suit if we can create a shortage."""
        player = self.player
        manager = global_vars.manager
        my_trumps = player.unplayed_cards[player.trump_suit]
        partners_trumps = player.unplayed_cards[player.trump_suit]

        suit_scores = {suit_name: 0 for suit_name in SUITS}
        for suit in SUITS:
            if suit != player.trump_suit.name:
                my_cards = player.unplayed_cards[suit]
                partners_cards = player.partners_unplayed_cards[suit]
                if len(my_cards) != len(partners_cards):
                    if ((len(my_cards) > len(partners_cards) and
                            len(partners_trumps) < len(my_trumps)) or
                            (len(my_cards) < len(partners_cards) and
                             len(my_trumps) < len(partners_trumps))):

                        if controls[suit] - len(player.partners_suit_cards[suit]) <= 1:
                            suit_scores[suit] += self.SHORTAGE_SCORE
                            manager.set_suit_to_develop(player.declarer, Suit(suit))
        return [(suit_name, score) for suit_name, score in suit_scores.items()]



"""Third seat card play for defender."""

from termcolor import cprint

import inspect
from ..logger import log

from bridgeobjects import SUITS, Card, CARD_VALUES, Suit
from .player import Player
from .utilities import (other_suit_for_signals, get_suit_strength,
                        trick_card_values, card_values)

from .third_seat import ThirdSeat
from .defender_play import deduce_partner_void_in_trumps, get_hilo_signal_card
import bfgcardplay.src.global_variables as global_vars

MODULE_COLOUR = 'blue'

MAXIMUM_TRICKS = 13
TRUMP_VALUE_UPLIFT = 13


class ThirdSeatDefender(ThirdSeat):
    def __init__(self, player: Player):
        super().__init__(player)

    def selected_card(self) -> Card:
        """Return the card if the third seat."""

        # Variables set in third_seat.py
        player = self.player
        trick = self.trick
        seat = self.seat

        cards = self.cards

        # Void
        if not cards:
            return self._select_card_if_void()

        # Singleton
        if len(cards) == 1:
            return log(inspect.stack(), cards[0])

        self._update_manager()

        # win trick if possible
        card = self._winning_card()
        if card:
            return card

        # Play high
        card = self._play_high_card()
        if card:
            player.highest_card[seat][card.suit.name] = card
            return card

        # Unblock suit
        card = self._unblock_suit()
        if card:
            return card

        # Play highest honour if winner
        card = self._play_winning_honour()
        if card:
            return card

        # Play winner if possible
        if (not player.is_winner_defender(trick.cards[0], trick) and
                not trick.cards[0].is_honour and
                cards[0].value > trick.cards[1].value):
            return log(inspect.stack(), cards[0])

        # Doubleton - play the higher
        card = self._top_of_doubleton()
        if card:
            return card

        # signal attitude
        return self._signal_attitude()

    def _update_manager(self):
        player = self.player
        trick = self.trick
        manager = global_vars.manager

        if len(player.board.tricks) == 1:
            manager.signal_card[player.partner_seat] = trick.cards[0]

        if player.trump_suit and trick.suit == player.trump_suit:
            if deduce_partner_void_in_trumps(player):
                player.voids[player.partner_seat][player.trump_suit.name] = True

        suit_name = trick.suit.name
        if player.trick_number == 1:
            if trick.cards[0].is_honour:
                manager.set_like_dislike(player.partner_seat, suit_name, True)

    def _play_high_card(self) -> Card:
        player = self.player
        cards = self.cards
        values = trick_card_values(self.trick, player.trump_suit)
        if self.trick.cards[0].is_honour and cards[0].value - values[0] <= 2:
            return None
        if cards[0].value > values[0] and cards[0].value > values[1]:
            return log(inspect.stack(), cards[0])
        return None

    def _unblock_suit(self) -> Card:
        player = self.player
        trick = self.trick
        cards = self.cards
        if player.trick_number == 1:
            if trick.cards[0].is_honour and len(cards) <= 2:
                return log(inspect.stack(), cards[0])
        if player.trick_number == 2:
            if player.board.tricks[0] == trick.suit:
                if player.board.tricks[0].cards[0].is_honour and len(cards) <= 2:
                    return log(inspect.stack(), cards[0])
        if len(cards) == 3:
            leading_card = trick.cards[0]
            if leading_card.value < cards[0].value:
                return log(inspect.stack(), cards[0])
        return None

    def _play_winning_honour(self) -> Card:
        player = self.player
        trick = self.trick
        # TODO is this too specific?
        suit = trick.suit
        jack = Card('J', suit.name)
        queen = Card('Q', suit.name)
        king = Card('K', suit.name)
        ace = Card('A', suit.name)
        if (trick.cards[0] == jack and
                ace in player.dummys_unplayed_cards[suit.name] and
                player.dummy_on_right and
                king in player.unplayed_cards[suit.name] and
                queen not in player.unplayed_cards[suit.name]):
            return log(inspect.stack(), king)
        return None

    def _top_of_doubleton(self) -> Card:
        player = self.player
        cards = self.cards
        manager = global_vars.manager
        suit_name = self.trick.suit.name
        if len(cards) == 2 and cards[0].value < CARD_VALUES['K']:
            manager.set_like_dislike(player.seat, suit_name, 1)
            return log(inspect.stack(), cards[0])
        return None

    def _signal_attitude(self) -> Card:
        player = self.player
        cards = self.cards
        manager = global_vars.manager
        suit_name = self.trick.suit.name
        if not manager.like_dislike(player.seat, suit_name) or not player.suit_rounds[suit_name]:
            if cards[0].is_honour:
                for card in cards[1:]:
                    if not card.is_honour:
                        if card.value >= CARD_VALUES['7']:
                            manager.set_like_dislike(player.seat, card.suit.name, 1)
                        return log(inspect.stack(), card)
                manager.set_like_dislike(player.seat, cards[-1].suit.name, 0)
        return get_hilo_signal_card(self)

    def _winning_card(self) -> Card | None:
        """Return the card if can win trick."""

        # No cards in trick suit, look for trump winner
        card = self._trump_if_void()
        if card:
            return card

        # Defeat contract if possible
        card = self._defeat_contract()
        if card:
            return card

        # Play winner if long suit
        card = self._winner_in_long_suit()
        if card:
            return card

        # Look for winner
        card = self._look_for_winner()
        if card:
            return card

        card = self._select_card_based_on_position()
        if card:
            return card

        return None

    def _trump_if_void(self) -> Card:
        if self.cards:
            return None
        player = self.player
        values = trick_card_values(self.trick, player.trump_suit)
        if player.trump and player.trump_cards:
            for card in player.trump_cards[::-1]:
                if (card.value + TRUMP_VALUE_UPLIFT > values[0] + 1 and
                        card.value + TRUMP_VALUE_UPLIFT > values[1]):
                    return log(inspect.stack(), card)
        return None

    def _defeat_contract(self) -> Card:
        player = self.player
        trick = self.trick
        if (player.is_winner_defender(self.cards[0], trick) and
                not player.is_winner_defender(trick.cards[0], trick) and
                player.defenders_tricks >= MAXIMUM_TRICKS - player.declarers_target):
            return log(inspect.stack(), self.cards[0])
        return None

    def _winner_in_long_suit(self) -> Card:
        player = self.player
        trick = self.trick
        cards = self.cards

        # Play winner if all winners
        card = self._all_winners_in_long_suit()
        if card:
            return card

        if len(cards) >= 5:
            for card in cards[::-1]:
                if player.is_winner_defender(trick.cards[0], trick):
                    return log(inspect.stack(), cards[-1])

                if player.is_winner_defender(card, trick):
                    return log(inspect.stack(), card)
        return None

    def _all_winners_in_long_suit(self) -> Card | None:
        # Play winner if all winners
        player = self.player
        trick = self.trick
        cards = self.cards

        winners = False
        dummys_cards = player.dummys_unplayed_cards[trick.suit.name]
        total_unplayed_cards = player.total_unplayed_cards[trick.suit.name]
        if len(cards) + len(dummys_cards) == len(total_unplayed_cards):
            winners = True
            for index, card in enumerate(dummys_cards):
                if index < len(self.cards):
                    if card.value > self.cards[index].value:
                        winners = False
                        break
        if winners:
            return log(inspect.stack(), cards[0])
        return None

    def _look_for_winner(self) -> Card:
        player = self.player
        trick = self.trick
        if not player.is_winner_defender(trick.cards[0], trick):
            top_touching_honour = player.touching_honours_in_hand(player.hand, trick.suit.name)
            if top_touching_honour and top_touching_honour.value > trick.cards[1].value:
                return log(inspect.stack(), top_touching_honour)
        return None

    def _select_card_based_on_position(self) -> Card:
        player = self.player
        if not player.dummy_on_left:
            return None

        cards = self.cards
        trick = self.trick
        # Win trick if possible
        my_values = card_values(cards)
        for index, value in enumerate(my_values[::-1]):
            if value > trick.cards[1].value:
                return log(inspect.stack(), cards[::-1][index])

        my_short_cards = [card for card in cards]
        for index, test_card in enumerate(my_short_cards[::-1]):
            # If dummy on left and can win, do so
            card = self._dummy_on_left(test_card)
            if card:
                return card

            card = self._dummy_tenace_not_dominated(test_card, index)
            if card:
                return card
        return None

    def _get_card_value(self, card) -> int:
        card_value = card.value
        # trick card values already adjusted for trumps
        if card.suit == self.player.trump_suit:
            card_value += TRUMP_VALUE_UPLIFT
        return card_value

    def _dummy_on_left(self, card) -> Card:
        player = self.player
        trick = self.trick

        if player.dummys_unplayed_cards[trick.suit.name]:
            values = trick_card_values(trick, player.trump_suit)
            if values[1] > values[0]:
                card_value = self._get_card_value(card)
                if card_value > player.dummys_unplayed_cards[trick.suit.name][0].value:
                    return log(inspect.stack(), card)
        return None

    def _dummy_tenace_not_dominated(self, card, index) -> Card:
        player = self.player
        card_value = self._get_card_value(card)
        values = trick_card_values(self.trick, player.trump_suit)
        if (card_value > values[0] + 3 and
                card_value > values[1] and
                len(self.cards) > index + 1 and
                card.value != self.cards[index+1].value + 1):
            if (not self._seat_dominates_left_hand_dummy_tenace(card) and
                    not self._ace_is_deprecated(self.trick, card)):
                return log(inspect.stack(), card)
        return None

    def _seat_dominates_left_hand_dummy_tenace(self, card: Card) -> bool:
        """Return True if hand dominated dummies tenace in that suit."""
        if self.player.dummy_on_left:
            return False
        tenace = self.player.dummy_suit_tenaces[card.suit.name]
        if tenace:
            if card.value > tenace.value:
                return True
        return False

    def _select_card_if_void(self) -> Card:
        """Return card if cannot follow suit."""
        player = self.player
        trick = self.trick
        # Trump if appropriate
        if player.trump_suit:
            values = trick_card_values(trick, player.trump_suit)
            if player.trump_cards:
                unplayed_cards = player.total_unplayed_cards[trick.suit.name]
                if unplayed_cards:
                    if (values[1] > values[0] or
                            trick.cards[0].value < unplayed_cards[0].value):
                        return log(inspect.stack(), player.trump_cards[-1])

        # Signal suit preference first time it is led."""
        if len(player.board.tricks) == 1:
            signal_card = self._signal_on_first_lead()
            if signal_card:
                return signal_card

        best_suit = self._best_suit(player)
        other_suit = other_suit_for_signals(best_suit)
        if other_suit != player.trump_suit:
            other_suit_cards = player.suit_cards[other_suit]
            if other_suit_cards and not other_suit_cards[-1].is_honour:
                return log(inspect.stack(), other_suit_cards[-1])

        long_suit_cards = {}
        selected_card = None
        for suit in SUITS:
            cards = player.suit_cards[suit]
            long_suit_cards[suit] = len(cards)
            if player.trump_suit and suit != player.trump_suit.name:
                if cards and not cards[-1].is_honour:
                    selected_card = cards[-1]
        if selected_card:
            return log(inspect.stack(), selected_card)

        for suit_name in SUITS:
            cards = player.unplayed_cards[suit]
            dummys_cards = player.dummys_unplayed_cards[suit]
            if len(cards) > len(dummys_cards):
                return log(inspect.stack(), cards[-1])

            # if suit_name != best_suit.name and suit_name != other_suit:
            #     final_suit_cards = player.suit_cards[suit_name]
            #     if final_suit_cards:
            #         return log(inspect.stack(), final_suit_cards[-1])

        # print(f'{player.suit_cards[suit][0]=}')
        max_length = 0
        for suit in SUITS:
            if long_suit_cards[suit] > max_length:
                max_length = long_suit_cards[suit]
                long_suit = suit
        return log(inspect.stack(), player.suit_cards[long_suit][-1])

    def _signal_on_first_lead(self) -> Card | None:
        """Return a card if it is first time that partner led it."""
        player = self.player
        trick = self.trick
        suits_already_signed = []
        if player.trump_suit:
            suits_already_signed.append(player.trump_suit)
        for board_trick in player.board.tricks:
            if board_trick.leader == player.partner_seat and board_trick != trick:
                suits_already_signed.append(board_trick.start_suit)

        if trick.start_suit not in suits_already_signed:
            suit = self._best_suit(player)
            suit_cards = player.suit_cards[suit.name]
            for card in suit_cards:
                if not card.is_honour:
                    return log(inspect.stack(), card)
        return None

    def _best_suit(self, player: Player) -> Suit:
        """Select suit for signal."""
        # TODO handle no points and equal suits
        cards = player.hand_cards.list
        suit_points = get_suit_strength(cards)
        max_points = 0
        best_suit = None
        for suit in SUITS:
            if player.trump_suit and suit == player.trump_suit.name:
                continue
            hcp = suit_points[suit]
            if hcp > max_points:
                max_points = hcp
                best_suit = suit
        if not best_suit:
            return player.longest_suit
        return Suit(best_suit)



""" Opening leads for Card Player class."""

import random
from termcolor import cprint

import inspect
from ..logger import log
from bridgeobjects import SUITS, Spades, Hearts, Diamonds, Clubs, SEATS, Denomination, Call

from .card_player_components import SelectedSuit, Card
import bfgcardplay.src.global_variables as global_vars


MODULE_COLOUR = 'cyan'


def opening_lead_suit(board):
    """Return the opening lead from a list of cards given the contract."""
    return OpeningLeadSuit(board).suit


class OpeningLeadSuit():
    """Generate a suit for the opening lead."""
    def __init__(self, board, *args, **kwargs):
        board.declarer = board.contract.declarer
        declarer_index = SEATS.index(board.declarer)
        seat_index = (declarer_index + 1) % 4
        self.seat = SEATS[seat_index]
        self.board = board
        self.contract = board.contract
        self.hand = board.hands[board.contract.leader]
        (leaders_bids, partners_bids, declarers_bids,
            dummys_bids, opponents_bids) = self._allocate_bids()
        self.leaders_bids = leaders_bids
        self.partners_bids = partners_bids
        self.declarers_bids = declarers_bids
        self.dummys_bids = dummys_bids
        self.opponents_bids = opponents_bids
        self.denomination = self.contract.denomination
        self.opponents_suits = self._opponents_suits()
        self.unbid_suits = self._unbid_suits()

        # Generate the resulting suit
        self.suit = self._opening_lead_suit()

    def _allocate_bids(self):
        """Return partner's bids and opponent's bids."""
        board = self.board
        auction, dealer, leader = board.auction, board.dealer, board.contract.leader
        dealer_index = SEATS.index(dealer)
        leader_index = SEATS.index(leader)
        dummy_index = (leader_index + 1) % 4
        partner_index = (leader_index + 2) % 4
        declarer_index = (leader_index + 3) % 4
        leaders_bids = []
        partners_bids = []
        declarers_bids = []
        dummys_bids = []
        opponents_bids = []
        opponents_dealt = (SEATS.index(dealer) - SEATS.index(leader)) % 2
        for index, call in enumerate(auction.calls):
            miss_call = 99
            if self.contract.trump_suit:
                if call == Call('4NT') or call == Call('5NT'):
                    miss_call = index + 2

            if call.is_value_call and not index == miss_call:
                if (index + opponents_dealt) % 2:
                    opponents_bids.append(call)
                if not call.is_pass and (dummy_index - dealer_index - index) % 4 == 0:
                    dummys_bids.append(call)
                elif not call.is_pass and (declarer_index - dealer_index - index) % 4 == 0:
                    declarers_bids.append(call)

            if not call.is_pass and (partner_index - dealer_index - index) % 4 == 0:
                partners_bids.append(call)
            elif not call.is_pass and (leader_index - dealer_index - index) % 4 == 0:
                leaders_bids.append(call)
        return (leaders_bids, partners_bids, declarers_bids, dummys_bids, opponents_bids)

    def _opening_lead_suit(self):
        """Return the opening lead."""
        if self.denomination.is_nt:
            return self._select_suit_for_nt_contract()
        else:
            assert self.denomination, 'No trump suit defined for suit contract.'
            return self._select_suit_for_suit_contract()

    def _select_suit_for_nt_contract(self):
        """Return the selected suit for a NT contract."""
        if self.contract.level >= 6:
            return self._select_suit_for_nt_slam_contract()

        if self.opponents_bids and self.opponents_bids[0].level >= 3:
            return self._select_suit_for_preempt_nt_contract()

        blind_opps = self._is_suit_call(self.opponents_bids)
        blind_partner = self._is_suit_call(self.partners_bids)
        if blind_opps and blind_partner:
            return self._select_suit_for_blind_nt_contract()

        return self._select_suit_for_informed_nt_contract()

    @staticmethod
    def _is_suit_call(calls):
        for call in calls:
            if call.is_suit_call:
                return False
        return True

    def _opponents_suits(self):
        """Return a list of suits in opponents_bids."""
        if (len(self.opponents_bids) > 1 and
                self.opponents_bids[0].is_nt and
                self.opponents_bids[1].denomination == Clubs()):
            stayman_suit = self._stayman_suits()
            if stayman_suit:
                return stayman_suit
            return [Clubs()]
        else:
            opponents_suits = []
            for bid in self.opponents_bids:
                if bid.denomination not in opponents_suits:
                    opponents_suits.append(bid.denomination)
            return opponents_suits

    def _stayman_suits(self):
        """Return the suits implied by stayman bids."""
        if len(self.opponents_bids) < 3:
            return None

        if self.opponents_bids[2].denomination == Diamonds():
            return [Hearts(), Spades()]

        if len(self.opponents_bids) < 4:
            return None
        if self.opponents_bids[2].denomination == Hearts() and self.opponents_bids[3].is_nt:
            return [Hearts(), Spades()]
        if self.opponents_bids[2].denomination == Spades() and self.opponents_bids[3].is_nt:
            return [Hearts(), Spades()]
        if self.opponents_bids[2].denomination == Spades():
            return [Spades()]
        if self.opponents_bids[2].denomination == Hearts():
            return [Hearts()]

    def _unbid_suits(self):
        """Return a list of unbid majors given opponents_bids."""
        unbid_suits = [suit for suit in SUITS]
        for bid in self.opponents_bids:
            if bid.denomination in unbid_suits:
                unbid_suits.remove(bid.denomination)
        return unbid_suits

    def _select_suit_for_nt_slam_contract(self):
        """Return the selected suit for an NT slam contract."""
        sequences = self.hand.honour_sequences()
        for suit in SUITS:
            if sequences[suit]:
                return log(inspect.stack(), SelectedSuit(suit, '002'))
            if self.hand.suit_holding[suit] >= 3 and self.hand.honours[suit] == 0:
                return log(inspect.stack(), SelectedSuit(suit, '022'))
        return self._select_suit_for_blind_nt_contract()

    def _select_suit_for_preempt_nt_contract(self):
        """Return the selected suit for an NT contract after preempt."""
        # Select suit with most High Card Points
        high_card_points = 0
        best_suit = None
        for suit in SUITS:
            suit_high_card_points = self.hand.suit_points(suit)
            if suit_high_card_points > high_card_points:
                high_card_points = suit_high_card_points
                best_suit = suit
        if best_suit:
            return log(inspect.stack(), SelectedSuit(best_suit, '021'))
        return log(inspect.stack(), SelectedSuit(self.hand.longest_suit.name, '021'))

    def _select_suit_for_informed_nt_contract(self):
        """Return the selected suit for an NT contract (not blind)."""
        manager = global_vars.manager
        if (self.hand.shape[0] >= 5 and self._has_entries() and
                self.hand.longest_suit not in self.opponents_suits):
            return log(inspect.stack(), SelectedSuit(self.hand.longest_suit.name, '021'))

        partners_suit = self._suit_bid_by_partner()
        suit_cards = [card for card in self.hand.cards if card.suit == partners_suit]
        if partners_suit and suit_cards:
            return log(inspect.stack(), partners_suit)

        sequences = self.hand.honour_sequences()
        max_sequence_length = self._max_sequence_length(sequences)
        long_sequences = self._long_suit_sequences(sequences, max_sequence_length)
        if ((self.hand.shape[0] >= 5 or max_sequence_length >= 4) and
                self.hand.longest_suit in long_sequences and
                self.hand.longest_suit not in self.opponents_suits):
            manager.working_suit[self.seat] = self.hand.longest_suit
            return log(inspect.stack(), SelectedSuit(self.hand.longest_suit.name, '021'))

        # Look for 4+ card unbid major
        four_card_major = self._biddable_major(4)
        if four_card_major:
            return log(inspect.stack(), four_card_major)

        # Biddable long suit
        if self.hand.shape[0] >= 5 and self.hand.longest_suit not in self.opponents_suits:
            return log(inspect.stack(), self.hand.longest_suit)

        # Look for 4 card suit
        four_card_suit = self._biddable_suit(4)
        if four_card_suit:
            return log(inspect.stack(), four_card_suit)

        # Best 3 card major
        three_card_major = self._biddable_major(3)
        if three_card_major:
            return log(inspect.stack(), three_card_major)

        # Look for 3 card suit
        three_card_suit = self._biddable_suit(3)
        if three_card_suit:
            return log(inspect.stack(), three_card_suit)

        return log(inspect.stack(), self.hand.longest_suit)

    def _biddable_major(self, suit_length):
        """Return longest biddable major based on suit length."""
        unbid_majors = []
        if self.hand.spades >= suit_length and Spades() not in self.opponents_suits:
            unbid_majors.append(Spades())
        if self.hand.hearts >= suit_length and Hearts() not in self.opponents_suits:
            unbid_majors.append(Hearts())
        if len(unbid_majors) > 1:
            if self.hand.spades > self.hand.hearts:
                return SelectedSuit('S', '019')
            if self.hand.hearts > self.hand.spades:
                return SelectedSuit('H', '019')
            # Select better suit
            if self.hand.honours['S'] > self.hand.honours['H'] and self.hand.honours['S'] > 1:
                return SelectedSuit('S', '019')
            if self.hand.honours['H'] > self.hand.honours['S'] and self.hand.honours['H'] > 1:
                return SelectedSuit('H', '019')
        if unbid_majors:
            suit_name = unbid_majors[0].name
            if self.hand.honours[suit_name] != 1:
                return SelectedSuit(suit_name, '019')
        return None

    def _biddable_suit(self, suit_length):
        """Return longest biddable suit based on suit length."""
        unbid_suits = []
        if self.hand.spades == suit_length and Spades() not in self.opponents_suits:
            unbid_suits.append(Spades())
        if self.hand.hearts == suit_length and Hearts() not in self.opponents_suits:
            unbid_suits.append(Hearts())
        if self.hand.diamonds == suit_length and Diamonds() not in self.opponents_suits:
            unbid_suits.append(Diamonds())
        if self.hand.clubs == suit_length and Clubs() not in self.opponents_suits:
            unbid_suits.append(Clubs())
        if len(unbid_suits) > 1:
            # Select best suit
            if Spades() in unbid_suits and self.hand.honours['S'] > 1:
                if (self.hand.honours['S'] > self.hand.honours['H'] and
                        self.hand.honours['S'] > self.hand.honours['D'] and
                        self.hand.honours['S'] > self.hand.honours['C']):
                    return SelectedSuit('S', '019')
            if Hearts() in unbid_suits and self.hand.honours['H'] > 1:
                if (self.hand.honours['H'] > self.hand.honours['D'] and
                        self.hand.honours['H'] > self.hand.honours['C']):
                    return SelectedSuit('H', '019')
            if Diamonds() in unbid_suits and self.hand.honours['D'] > 1:
                if (self.hand.honours['D'] > self.hand.honours['C']):
                    return SelectedSuit('D', '019')
            if Clubs() in unbid_suits and self.hand.honours['C'] > 1:
                return SelectedSuit('C', '019')
        if unbid_suits:
            suit_name = unbid_suits[0].name
            return SelectedSuit(suit_name, '019')
        return None

    def _has_entries(self):
        """Return True if the hand has outside entries."""
        # print('_has_entries')
        suit_points = 0
        for suit in SUITS:
            if suit != self.hand.longest_suit:
                suit_points += self.hand.suit_points(suit)
        if suit_points >= 7:
            return True
        return False

    def _select_suit_for_blind_nt_contract(self):
        """Return the selected suit for a blind NT contract."""

        if self.hand.shape[0] >= 5:
            return self._select_long_suit_for_blind_nt_contract()

        return self._select_four_card_suit_for_blind_nt_contract()

    def _select_long_suit_for_blind_nt_contract(self):
        """Return the selected long suit for a blind NT contract."""
        # honour sequences
        best_suit_with_sequence = self._best_suit_with_sequence()

        # the number of honours in the longest suit
        long_suit_honours = self.hand.honours[self.hand.longest_suit.name]

        # Select the best long suit with a 3 card sequence.
        if best_suit_with_sequence:
            suit = best_suit_with_sequence
            suit.selection_reason = '002'
            return log(inspect.stack(), suit)

        # Select a long suit with multiple honours.
        if self.hand.shape[0] >= 5 and long_suit_honours >= 2:
            long_suit = self.hand.longest_suit.name
            return log(inspect.stack(), SelectedSuit(long_suit, selection_reason='001'))

        # Select from best 5+ card suit
        if self.hand.shape[0] == self.hand.shape[1]:
            long_suits = self._equal_long_suits()
            suit = self._select_from_equal_long_suits(long_suits)
            return suit

        suit = SelectedSuit(self.hand.longest_suit.name, selection_reason='003')
        return log(inspect.stack(), suit)

    def _select_four_card_suit_for_blind_nt_contract(self):
        """Return the selected 4 card suit for a blind NT contract."""
        # honour sequences
        best_suit_with_sequence = self._best_suit_with_sequence()
        best_suit_with_near_sequence = self._best_suit_with_near_sequence()
        best_suit_with_interior_sequence = self._best_suit_with_interior_sequence()

        # Select best 4 card suit with a sequence
        if best_suit_with_sequence:
            suit = best_suit_with_sequence
            suit.selection_reason = '010'
            return log(inspect.stack(), suit)

        # Select best 4 card suit with a  near sequence
        if best_suit_with_near_sequence:
            suit = best_suit_with_near_sequence
            suit.selection_reason = '011'
            return log(inspect.stack(), suit)

        # Select best 4 card suit with an interior sequence
        if best_suit_with_interior_sequence:
            suit = best_suit_with_interior_sequence
            suit.selection_reason = '012'
            return log(inspect.stack(), suit)

        # Select best 4 card suit
        if self.hand.shape[0] == self.hand.shape[1]:
            suit = self._select_best_four_card_suit()
            suit.selection_reason = '006'
            return log(inspect.stack(), suit)

        suit = SelectedSuit(self.hand.longest_suit.name, selection_reason='005')
        return log(inspect.stack(), suit)

    def _select_best_four_card_suit(self):
        """Select the better of two 4 card suits."""
        four_card_suits = self._equal_long_suits()

        # Reject suits with AQxx or KJxx
        if self.contract.trump_suit:
            four_card_suits = self._remove_tenaces(four_card_suits)

        if len(four_card_suits) == 1:
            return (four_card_suits[0], '010')

        if not four_card_suits:
            four_card_suits = self._equal_long_suits()
        suit = self._select_from_equal_long_suits(four_card_suits)
        return suit

    def _select_from_equal_long_suits(self, long_suits):
        """Select the best suit from equal length by number of honours."""
        # Create a list of suits with maximum length for the hand.
        long_suits = self._equal_long_suits()

        # Find if one suit is headed by an Ace and the other(s) are not.
        non_ace_suits = self._non_ace_suits(long_suits)
        if len(non_ace_suits) == 1:
            suit = non_ace_suits[0]
            suit.selection_reason = '007'
            return log(inspect.stack(), suit)

        elif len(non_ace_suits) > 1:
            long_suits = non_ace_suits

        # Find suits with touching honours
        touching_honours = self.hand.touching_honours()
        touching_honour_suits = []
        for suit in long_suits:
            if touching_honours[suit.name]:
                touching_honour_suits.append(suit)
        if len(touching_honour_suits) == 1:
            touching_honour_suits[0].selection_reason = '009'
            return log(inspect.stack(), touching_honour_suits[0])
        elif len(touching_honour_suits) >= 2:
            long_suits = touching_honour_suits

        # Find the maximum number of honours in the long suits.
        most_honours = self._most_honours(long_suits)

        # Create a list of long suits with that number of honours.
        suits_with_honours = self._suits_with_most_honours(long_suits, most_honours)

        # Select the suit with the best honours.
        if len(suits_with_honours) > 1:
            # Find the maximum honour points in long suits.
            suit = self._best_of_long_suits_with_honours(suits_with_honours)
        else:
            suit = suits_with_honours[0]
        suit.selection_reason = '004'
        return log(inspect.stack(), suit)

    def _remove_tenaces(self, candidate_suits):
        """Remove suits with tenaces from a list of suits."""
        candidates = [suit for suit in candidate_suits]
        for suit in candidates:
            if ((Card(f'A{suit.name}') in self.hand.cards and
                    Card(f'Q{suit.name}') in self.hand.cards and
                    Card(f'K{suit.name}') not in self.hand.cards) or
                    (Card(f'K{suit.name}') in self.hand.cards and
                     Card(f'J{suit.name}') in self.hand.cards and
                     Card(f'Q{suit.name}') not in self.hand.cards)):
                candidate_suits.remove(suit)
        return candidate_suits

    def _most_honours(self, suit_list):
        """Return the maximum number of honours in a list of suits."""
        most_honours = -1
        for suit in suit_list:
            honours = self.hand.honours[suit.name]
            if honours > most_honours:
                most_honours = honours
        return most_honours

    def _suits_with_most_honours(self, suit_list, most_honours):
        """Return a list of suits with the given number of honours."""
        long_suits = []
        for suit in suit_list:
            honours = self.hand.honours[suit.name]
            if honours == most_honours:
                long_suits.append(suit)
        return long_suits

    def _best_of_long_suits_with_honours(self, suit_list):
        """Return the best of suits with honours."""
        max_honour_points = self._most_honour_points(suit_list)
        honour_suits = self._suits_with_honour_points(suit_list, max_honour_points)
        if len(honour_suits) == 1:
            suit = SelectedSuit(honour_suits[0].name)
        else:
            max_value = self._highest_suit_value(honour_suits)
            value_suits = self._get_suits_with_value(honour_suits, max_value)
            suit = value_suits[0]
        suit.selection_reason = '008'
        return suit

    def _equal_long_suits(self):
        """Return a list of suits of equal long length."""
        long_suits = []
        for suit_name, suit in SUITS.items():
            if self.hand.suit_holding[suit_name] == self.hand.shape[0]:
                long_suits.append(SelectedSuit(suit.name))
        return long_suits

    def _most_honour_points(self, suit_list):
        """Return the maximum number of honour points in a list of suits."""
        max_honour_points = -1
        for suit in suit_list:
            if self.hand.suit_points(suit) >= max_honour_points:
                max_honour_points = self.hand.suit_points(suit)
        return max_honour_points

    def _suits_with_honour_points(self, suit_list, max_honour_points):
        """Return a list of suits with the given honour points."""
        honour_suits = []
        for suit in suit_list:
            if self.hand.suit_points(suit) == max_honour_points:
                honour_suits.append(suit)
        return honour_suits

    def _best_suit_with_sequence(self):
        """Return the suit sequence with the highest honour."""
        sequences = self.hand.honour_sequences()

        # find length of longest sequences
        max_sequence_length = self._max_sequence_length(sequences)
        if max_sequence_length == 0:
            return None

        # find suits with longest sequences in 4+ card suits
        long_sequences = self._long_suit_sequences(sequences, max_sequence_length)
        if len(long_sequences) == 0:
            return None
        if len(long_sequences) == 1:
            return long_sequences[0]

        # find longest suit with longest sequence
        long_suits = self._get_longest_suits_with_sequence(long_sequences)
        if len(long_suits) == 1:
            return long_sequences[0]

        #  find highest ranking sequence
        ranking_suits = self._get_highest_ranking_sequence(long_suits, sequences)
        if ranking_suits and len(ranking_suits) == 1:
            return long_sequences[0]

        # return major if equal ranking suits
        if ranking_suits:
            for suit in ranking_suits:
                if suit.is_major:
                    return suit

            # more than one suit with equal credentials, so just return one of them
            return random.choice(ranking_suits)

        return None

    def _get_longest_suits_with_sequence(self, sequences):
        # find longest suit with sequence
        suit_length = self._get_length_of_long_suits_with_sequence(sequences)
        long_suits = []
        for suit in sequences:
            if self.hand.suit_holding[suit.name] == suit_length:
                long_suits.append(suit)
        return long_suits

    def _get_length_of_long_suits_with_sequence(self, sequences):
        # find longest suit with sequence
        suit_length = 0
        for suit in sequences:
            if self.hand.suit_holding[suit.name] > suit_length:
                suit_length = self.hand.suit_holding[suit.name]
        return suit_length

    def _get_highest_ranking_sequence(self, long_suits, sequences):
        top_value = self._get_highest_card_in_sequence(long_suits, sequences)
        ranking_suits = []
        for suit in long_suits:
            for card in sequences[suit.name]:
                if card.value == top_value:
                    ranking_suits.append(suit)

    def _get_highest_card_in_sequence(self, long_suits, sequences):
        top_value = 0
        for suit in long_suits:
            for card in sequences[suit.name]:
                if card.value > top_value:
                    top_value = card.value
        return top_value

    def _max_sequence_length(self, sequences):
        """Return the maximum sequence length (or zero)."""
        max_sequence_length = 0
        for suit_name, honours in sequences.items():
            if len(honours) > max_sequence_length:
                max_sequence_length = len(honours)
        return max_sequence_length

    def _long_suit_sequences(self, sequences, max_sequence_length):
        """Return the a list of 4+ card suits with sequences."""
        long_sequences = []
        for suit_name, honours in sequences.items():
            if len(honours) == max_sequence_length:
                if self.hand.suit_holding[suit_name] >= 4:
                    suit = SelectedSuit(suit_name)
                    suit.sequence = sequences[suit_name]
                    long_sequences.append(suit)
        return long_sequences

    def _best_suit_with_near_sequence(self):
        """Return the near sequence with the highest honour."""
        candidates = []
        for suit_name in SUITS:
            if self.hand.suit_holding[suit_name] == 4:
                candidates.append(suit_name)
        if not candidates:
            return None
        for suit_name in candidates:
            suit = SelectedSuit(suit_name)
            if (Card(f'K{suit_name}') in self.hand.cards and
                    Card(f'Q{suit_name}') in self.hand.cards and
                    Card(f'T{suit_name}') in self.hand.cards):
                return suit
            elif (Card(f'Q{suit_name}') in self.hand.cards and
                    Card(f'J{suit_name}') in self.hand.cards and
                    Card(f'9{suit_name}') in self.hand.cards):
                return suit
            elif (Card(f'J{suit_name}') in self.hand.cards and
                    Card(f'T{suit_name}') in self.hand.cards and
                    Card(f'8{suit_name}') in self.hand.cards):
                return suit
            elif (Card(f'T{suit_name}') in self.hand.cards and
                    Card(f'9{suit_name}') in self.hand.cards and
                    Card(f'7{suit_name}') in self.hand.cards):
                return suit
        return None

    def _best_suit_with_interior_sequence(self):
        """Return the interior sequence with the highest honour."""
        candidates = []
        for suit in SUITS:
            if self.hand.suit_holding[suit] == 4:
                candidates.append(suit)
        if not candidates:
            return None
        for suit_name in candidates:
            suit = SelectedSuit(suit_name)
            if (Card(f'A{suit_name}') in self.hand.cards and
                    Card(f'Q{suit_name}') in self.hand.cards and
                    Card(f'J{suit_name}') in self.hand.cards):
                return suit
            elif (Card(f'K{suit_name}') in self.hand.cards and
                    Card(f'J{suit_name}') in self.hand.cards and
                    Card(f'T{suit_name}') in self.hand.cards):
                return suit
            elif (Card(f'Q{suit_name}') in self.hand.cards and
                    Card(f'T{suit_name}') in self.hand.cards and
                    Card(f'9{suit_name}') in self.hand.cards):
                return suit
            elif (Card(f'J{suit_name}') in self.hand.cards and
                    Card(f'9{suit_name}') in self.hand.cards and
                    Card(f'8{suit_name}') in self.hand.cards):
                return suit
            elif (Card(f'T{suit_name}') in self.hand.cards and
                    Card(f'8{suit_name}') in self.hand.cards and
                    Card(f'7{suit_name}') in self.hand.cards):
                return suit
        return None

    def _non_ace_suits(self, suit_list):
        """Return a list of suits with no ace if there is a long suit with an ace."""
        ace_suits = []
        non_ace_suits = []
        for suit in suit_list:
            card = Card(f'A{suit.name}')
            if card in self.hand.cards:
                ace_suits.append(suit)
            else:
                non_ace_suits.append(suit)
        if len(ace_suits) >= 1:
            return non_ace_suits
        else:
            return []

    def _highest_suit_value(self, suit_list):
        """Return the maximum suit value for a list of suits."""
        max_value = -1
        for suit in suit_list:
            cards = self.hand.cards_by_suit[suit.name]
            value = 0
            for card in cards:
                value += card.value
            if value > max_value:
                max_value = value
        return max_value

    def _get_suits_with_value(self, suit_list, max_value):
        """Return a list of suits with max_value."""
        value_suits = []
        for suit in suit_list:
            cards = self.hand.cards_by_suit[suit.name]
            value = 0
            for card in cards:
                value += card.value
            if value == max_value:
                value_suits.append(suit)
        return value_suits

    def _select_suit_for_suit_contract(self):
        """Return the selected suit for a suit contract."""
        # Order of selection based on Klinger's "Basic Bridge" p. 138
        selected_suit = None

        selection_functions = [
            self._defend_against_slam,
            self._trump_after_double,
            self._dummy_has_long_suit,
            self._suit_with_AK,
            self._suit_with_sequence,
            self._suit_with_singleton,
            self._suit_bid_by_partner,
            self._partners_shortage,
            self._best_suit_with_near_sequence,
            self._best_suit_with_interior_sequence,
            self._use_trumps,
            self._touching_honours,
            self._suit_if_not_deprecated,
            self._best_of_poor_suits,
        ]

        for selection_function in selection_functions:
            if self._no_selected_suit_or_no_cards(selected_suit):
                selected_suit = selection_function()
            else:
                break

        if self._no_selected_suit_or_no_cards(selected_suit):
            print('No suit found.', self.hand, self.denomination, self.opponents_bids)
            assert False, 'No suit found for suit contract.'
        return selected_suit

    def _defend_against_slam(self):
        """Return a suit against a slam bid."""
        if self.contract.level >= 6:
            if Call('D') in self.partners_bids:
                calls = [call for call in self.board.auction.calls]
                calls.reverse()
                double = False
                for call in calls:
                    if double:
                        if call.is_suit_call:
                            return log(inspect.stack(), SelectedSuit(call.denomination.name, '031'))
                        else:
                            for dummys_call in self.dummys_bids:
                                if dummys_call.is_suit_call:
                                    dummys_suit = dummys_call.denomination.name
                                    return log(inspect.stack(), SelectedSuit(dummys_suit, '032'))
                    if call == Call('D'):
                        double = True
            if self.hand.hcp >= 7:
                for suit in SUITS:
                    if suit != self.contract.denomination.name:
                        if Card(f'K{suit}') in self.hand.cards:
                            return log(inspect.stack(), SelectedSuit(suit, '030'))
            return None
        return None

    def _no_selected_suit_or_no_cards(self, selected_suit):
        """Return True if a suit has not been selected or the hand has no cards in that suit."""
        if not selected_suit:
            return True
        elif self.hand.cards_in_suit(selected_suit) == 0:
            return True
        return False

    def _suit_with_sequence(self):
        """Return the best suit with a sequence if any."""
        sequences = self.hand.honour_sequences()
        selected_suit = self._get_best_sequence_suit(sequences)
        if selected_suit:
            selected_suit.selection_reason = '002'
        return selected_suit

    def _trump_after_double(self):
        """Return the trump suit if partner has passed your takeout double."""
        partner_passed = True
        for call in self.partners_bids:
            if not call.is_pass:
                partner_passed = False
                break
        if Call('D') in self.leaders_bids and partner_passed:
            return log(inspect.stack(), SelectedSuit(self.contract.denomination.name, '028'))

    def _dummy_has_long_suit(self):
        """Return the unbid suit if dummy has long suit."""
        if len(self.board.tricks) == 1:
            return None
        if len(self.dummys_bids) < 2:
            return None
        denominations = {}
        for call in self.dummys_bids:
            if call.denomination in denominations:
                denominations[call.denomination] += 1
            else:
                denominations[call.denomination] = 0
        for denomination, value in denominations.items():
            if value:
                suit = self._equal_long_suits()[0]
                return log(inspect.stack(), SelectedSuit(suit.name, '029'))
        return None

    def _suit_with_AK(self):
        """Return suit with an AK pairing."""
        candidate_suits = []
        for suit in SUITS:
            if suit != self.denomination.name:
                cards = self.hand.cards_by_suit[suit]
                if len(cards) >= 2:
                    if cards[0].rank == 'A' and cards[1].rank == 'K':
                        candidate_suits.append(suit)
        if len(candidate_suits) > 1:
            partners_suits = [call.denomination.name for call in self.partners_bids]
            for suit in candidate_suits:
                if suit in partners_suits:
                    return log(inspect.stack(), SelectedSuit(suit, '006'))
        elif candidate_suits:
            return log(inspect.stack(), SelectedSuit(candidate_suits[0], '007'))
        return None

    def _suit_with_singleton(self):
        """Return suit with a singleton."""
        trump_cards = self.hand.cards_by_suit[self.contract.name[1]]
        trump_hcp = 0
        for card in trump_cards:
            trump_hcp += card.high_card_points
        if trump_hcp >= 4 and len(trump_cards) >= 3:
            return None
        candidate_suits = []
        for suit_name, suit in SUITS.items():
            if suit != self.denomination:
                cards = self.hand.cards_by_suit[suit_name]
                if len(cards) == 1:
                    candidate_suits.append(suit)
        if candidate_suits:
            return log(inspect.stack(), SelectedSuit(candidate_suits[0].name, '008'))
        return None

    def _suit_bid_by_partner(self):
        """Return suit bid by partner."""
        for call in self.partners_bids:
            if call.is_suit_call:
                return log(inspect.stack(), SelectedSuit(call.denomination.name, '020'))
        return None

    def _suit_if_not_deprecated(self):
        """Return the best suit if not deprecated."""
        candidate_suits = self._non_deprecated_suits()
        if not candidate_suits:
            return None

        suit = self._suit_from_non_deprecated_suits(candidate_suits)
        if suit:
            return suit

        # Select the longest remaining suit
        max_length = max([self.hand.cards_by_suit[suit] for suit in candidate_suits])
        for suit in candidate_suits:
            cards = self.hand.cards_by_suit[suit]
            if len(cards) == max_length:
                return log(inspect.stack(), SelectedSuit(suit, '018'))
        return None

    def _suit_from_non_deprecated_suits(self, suits):
        if len(suits) == 1:
            return log(inspect.stack(), SelectedSuit(suits[0], '014'))

        for suit in suits:
            cards = self.hand.cards_by_suit[suit]
            honour = self._cards_contain_honour(cards)

            # Select suit with two honours
            if self._suit_has_two_honours(suit, cards):
                return log(inspect.stack(), SelectedSuit(suit, '015'))

            # Select doubleton with no honour
            if len(cards) == 2 and not honour:
                return log(inspect.stack(), SelectedSuit(suit, '016'))

            # Select suit with no honours
            if not honour:
                return log(inspect.stack(), SelectedSuit(suit, '017'))
        return None

    def _cards_contain_honour(self, cards):
        for card in cards:
            if card.is_honour:
                return True
        return False

    def _suit_has_two_honours(self, suit, cards):
        """Return True if the suit has at least two honours."""
        if Card(f'K{suit}') in cards and Card(f'Q{suit}') in cards:
            return True
        elif Card(f'K{suit}') in cards and Card(f'J{suit}') in cards:
            return True
        elif Card(f'K{suit}') in cards and Card(f'T{suit}') in cards:
            return True
        elif Card(f'Q{suit}') in cards and Card(f'J{suit}') in cards:
            return True
        elif Card(f'Q{suit}') in cards and Card(f'T{suit}') in cards:
            return True
        return False

    def _non_deprecated_suits(self):
        """Return a list of non-deprecated suits."""
        opposition_suits = [
            call.denomination.name
            for call in self.opponents_bids
            if call.is_suit_call
        ]
        candidate_suits = [suit for suit in SUITS]
        for suit in SUITS:
            remove = False
            cards = self.hand.cards_by_suit[suit]

            # Do not underlead an A
            if len(cards) >= 2 and cards[0].rank == 'A' and cards[1].rank != 'K':
                remove = True

            # Do not lead from a doubleton headed by an honour
            elif (len(cards) == 2 and cards[0].high_card_points and
                  cards[1].value <= Card('9S').value):
                remove = True

            # Do not lead a singleton trump
            elif suit == self.denomination.name and len(cards) == 1:
                remove = True

            # Ignore suits bid by opposition
            elif suit in opposition_suits:
                remove = True

            if remove:
                candidate_suits.remove(suit)
        return candidate_suits

    def _use_trumps(self):
        """Return the trump suit if it can be used."""
        cards = self.hand.cards_by_suit[self.denomination.name]
        if len(cards) >= 2 and not cards[0].high_card_points:
            return log(inspect.stack(), SelectedSuit(self.denomination.name, '028'))
        return None

    def _get_best_sequence_suit(self, sequences):
        """Return the best suit sequence."""
        top_of_sequence = self._get_top_card_in_sequences(sequences)
        candidates = self._get_candidate_sequence_suits(sequences, top_of_sequence)
        if len(candidates) == 1:
            return log(inspect.stack(), SelectedSuit(candidates[0], '001'))
        return self._get_best_suit_for_equal_rank_sequences(candidates)

    def _get_top_card_in_sequences(self, sequences):
        """Return the highest value card from a set of sequences."""
        top_of_sequence = 0
        for suit, sequence in sequences.items():
            if sequence:
                value = sequence[0].value
                if value > top_of_sequence:
                    top_of_sequence = value
        return top_of_sequence

    def _get_candidate_sequence_suits(self, sequences, top_of_sequence):
        """Return a list of suit sequences with equal top card."""
        length_of_sequence = 0
        candidates = []
        for suit, sequence in sequences.items():
            if suit != self.denomination.name:
                if sequence:
                    value = sequence[0].value
                    if value == top_of_sequence:
                        if len(sequence) >= length_of_sequence:
                            length_of_sequence = len(sequence)
                            candidates.append(suit)
        return candidates

    def _get_best_suit_for_equal_rank_sequences(self, candidate_suits):
        """Return the suit with the maximum length in sequences."""
        max_suit_length = 0
        suit_candidates = []
        for suit in candidate_suits:
            if self.hand.suit_holding[SelectedSuit(suit)] > max_suit_length:
                max_suit_length = self.hand.suit_holding[SelectedSuit(suit)]
        for suit in candidate_suits:
            if self.hand.suit_holding[SelectedSuit(suit)] == max_suit_length:
                suit_candidates.append(suit)
        if len(suit_candidates) == 1:
            return SelectedSuit(suit_candidates[0], '002')
        else:
            majors = []
            for suit in suit_candidates:
                if suit in 'SH':
                    majors.append(suit)
                if len(majors) == 1:
                    return SelectedSuit(majors[0], '003')
                elif len(majors) == 2:
                    suit_candidates = majors
        if len(suit_candidates) > 1:
            suit_candidates = self._get_suit_with_highest_value(suit_candidates)
        return suit_candidates

    def _get_suit_with_highest_value(self, suit_candidates):
        """Return the suit with  the highest aggregate card value."""
        ranking = 0
        chosen_suit = None
        for suit in suit_candidates:
            values = 0
            for card in self.hand.cards_by_suit[suit]:
                values += card.value
            if values > ranking:
                ranking = values
                chosen_suit = suit
        return SelectedSuit(chosen_suit, '004')

    def _touching_honours(self):
        """Return suit with touching honours."""
        touching_honours = self.hand.touching_honours()
        for suit in SUITS:
            if len(touching_honours[suit]) >= 2:
                return log(inspect.stack(), SelectedSuit(suit, '027'))
        return None

    def _partners_shortage(self):
        """Return partners short suit."""
        last_suit = ''
        suit_fit = None
        for call in self.opponents_bids:
            if call.denomination == last_suit and call.denomination != self.contract.denomination:
                if call.is_suit_call:
                    suit_fit = call.denomination
                    break
            else:
                last_suit = call.denomination
        if suit_fit and len(self.hand.cards_by_suit[suit_fit.name]) >= 4:
            return log(inspect.stack(), SelectedSuit(suit_fit.name, '026'))
        return None

    def _best_of_poor_suits(self):
        """Select best suit from poor choices."""
        suits = [Denomination(suit) for suit in SUITS if Denomination(suit) not in self.opponents_suits]
        deprecated_suits = []
        for suit in suits:
            cards = self.hand.cards_by_suit[suit.name]
            if len(cards) >= 2:
                if cards[0].rank == 'A' and cards[1].rank != 'K':
                    deprecated_suits.append(suit)
        for suit in deprecated_suits:
            suits.remove(suit)

        if len(suits) == 1 and self.hand.cards_by_suit[suits[0].name]:
            return log(inspect.stack(), SelectedSuit(suits[0].name, '016'))
        # Use K high suit rather than Q etc.
        for suit in suits:
            cards = self.hand.cards_by_suit[suit.name]
            if Card(f'K{suit.name}') in cards:
                return log(inspect.stack(), SelectedSuit(suit.name, '023'))
            if Card(f'Q{suit.name}') in cards:
                return log(inspect.stack(), SelectedSuit(suit.name, '024'))
            if Card(f'J{suit.name}') in cards:
                return log(inspect.stack(), SelectedSuit(suit.name, '025'))

        cards = self.hand.cards_by_suit[self.hand.longest_suit.name]
        if cards[0].rank != 'A':
            selected_suit = SelectedSuit(self.hand.longest_suit.name, '005')
        else:
            selected_suit = SelectedSuit(self.hand.second_suit.name, '888')
        return selected_suit



from bridgeobjects import Trick, Card, Suit, parse_pbn, Board

from ..src.player import Player

from ..src.utilities import (get_seat, get_suit_strength, trick_card_values, get_winning_card,
                                other_suit_for_signals, get_list_of_best_scores,
                                highest_remaining_card, play_to_unblock, get_long_hand)


def test_get_seat():
    trick = Trick()
    trick.leader = 'W'
    seat = get_seat(trick, 3)
    assert seat == 'S'

    trick.leader = 'N'
    seat = get_seat(trick, 2)
    assert seat == 'S'


def test_get_suit_strength():
    cards = [Card('AS'), Card('JS'), Card('TS'), Card('QC'), Card('KD')]
    suit_strength = get_suit_strength(cards)
    assert suit_strength['S'] == 5
    assert suit_strength['H'] == 0
    assert suit_strength['D'] == 3
    assert suit_strength['C'] == 2


def test_other_suit_for_signals():
    assert other_suit_for_signals(Suit('S')) == 'C'
    assert other_suit_for_signals(Suit('H')) == 'D'
    assert other_suit_for_signals(Suit('D')) == 'H'
    assert other_suit_for_signals(Suit('C')) == 'S'


def test_get_list_of_best_scores_distinct():
    candidates = {Suit("C"): 8, Suit("D"): 6, Suit("H"): 5, Suit("S"): 7}
    best_candidates = get_list_of_best_scores(candidates)
    assert best_candidates == [Suit("C")]

    best_candidates = get_list_of_best_scores(candidates, True)
    assert best_candidates == [Suit("H")]


def test_get_list_of_best_scores_duplicate():
    candidates = {Suit("C"): 8, Suit("D"): 6, Suit("H"): 6, Suit("S"): 7}
    best_candidates = get_list_of_best_scores(candidates, True)
    assert Suit('H') in best_candidates
    assert Suit('D') in best_candidates
    assert Suit('S') not in best_candidates


def test_trick_card_values_empty():
    trick = Trick()
    values = trick_card_values(trick)
    assert values == [0, 0, 0, 0]


def test_trick_card_values_one():
    trick = Trick()
    trick.cards = [Card('3S')]
    values = trick_card_values(trick)
    assert values == [2]


def test_trick_card_values_two():
    trick = Trick()
    trick.cards = [Card('3S'), Card('4S')]
    values = trick_card_values(trick)
    assert values == [2, 3]


def test_trick_card_values_triple():
    trick = Trick()
    trick.cards = [Card('3S'), Card('4S'), Card('KS')]
    values = trick_card_values(trick)
    assert values == [2, 3, 12]


def test_trick_card_values_discard_no_trumps():
    trick = Trick()
    trick.cards = [Card('3S'), Card('4C'), Card('KS')]
    values = trick_card_values(trick)
    assert values == [2, 0, 12]


def test_trick_card_values_trumps():
    trick = Trick()
    trick.cards = [Card('3S'), Card('4S'), Card('KS'), Card('2S')]
    values = trick_card_values(trick, Suit('S'))
    assert values == [15, 16, 25, 14]


def test_trick_card_values_discard_trumps():
    trick = Trick()
    # Discard in trumps
    trick.cards = [Card('3S'), Card('4C'), Card('2D'), Card('KS')]
    values = trick_card_values(trick, Suit('C'))
    assert values == [2, 16, 0, 12]


def test_trick_card_values_discard_trumps_two():
    trick = Trick()
    trick.cards = [Card('3S'), Card('4C'), Card('KS'), Card('2D')]
    values = trick_card_values(trick, Suit('C'))
    assert values == [2, 16, 12, 0]


def test_get_winning_card_discard():
    trick = Trick()

    trick.cards = [Card('3S'), Card('4C'), Card('KS')]
    card = get_winning_card(trick, None)
    assert card == Card('KS')


def test_get_winning_card_trump():
    trick = Trick()

    trick.cards = [Card('3S'), Card('4C'), Card('KS')]
    card = get_winning_card(trick, Suit('C'))
    assert card == Card('4C')


def _generate_board(board_pbn: list) -> Board:
    board = parse_pbn(board_pbn)[0].boards[0]
    for seat, hand in board.hands.items():
        hand.unplayed_cards = [card for card in hand.cards]
    return board


def _generate_trick(leader: str, cards: list[Card]) -> Trick:
    trick = Trick()
    trick.leader = leader
    trick.cards = cards
    return trick


def test_highest_remaining_card():
    board_pbn = [
        '[Board "0"]',
        '[Description "20 Jun 2023 17:01:12"]',
        '[Dealer "S"]',
        '[Vulnerable "None"]',
        '[Deal "S:AJ2.AQJ7.AT6.QJ9 T6.2.KJ832.K853 Q743.8.Q7.AT764 K985.KT96543..2"]',
        '[Declarer S]',
        '[Contract 3NT]',
        '[Auction "S"]', '3NT Pass Pass Pass',
        '[Play "W"]'
    ]
    board = _generate_board(board_pbn)

    trick = _generate_trick('E', [Card('4D'), Card('5D'), Card('9D'), Card('TD')])
    board.tricks.append(trick)
    player = Player(board)
    card = highest_remaining_card(player, trick)
    assert card == Card('8D')

# When unlocking you need to test leading from both long and short hands

def test_play_to_unblock_high_in_short_from_long():
    hand_to_play = [Card('JS'), Card('3S'), Card('2S')]
    other_hand = [Card('KS'), Card('4S')]
    card = play_to_unblock(hand_to_play, other_hand)
    assert card == Card('2S')

def test_play_to_unblock_high_in_short_from_short():
    hand_to_play = [Card('KS'), Card('4S')]
    other_hand = [Card('JS'), Card('3S'), Card('2S')]
    card = play_to_unblock(hand_to_play, other_hand)
    assert card == Card('KS')


def test_play_to_unblock_singleton_in_short_from_long():
    hand_to_play = [Card("JD"), Card("8D"), Card("4D")]
    other_hand = [Card('7D')]
    card = play_to_unblock(hand_to_play, other_hand)
    assert card == Card('JD')


def test_play_to_unblock_singleton_in_short_from_short():
    hand_to_play = [Card('7D')]
    other_hand = [Card("JD"), Card("8D"), Card("4D")]
    card = play_to_unblock(hand_to_play, other_hand)
    assert card == Card('7D')


def test_play_to_unblock_void_in_short():
    hand_to_play = []
    other_hand = [Card("JD"), Card("8D"), Card("4D")]
    card = play_to_unblock(hand_to_play, other_hand)
    assert card == Card('JD')


def test_play_to_unblock_three_honours_from_long():
    hand_to_play = [Card("AD"), Card("KD"), Card("7D")]
    other_hand = [Card("QD"), Card("9D")]
    card = play_to_unblock(hand_to_play, other_hand)
    assert card == Card('7D')


def test_play_to_unblock_three_honours_from_short():
    hand_to_play = [Card("QD"), Card("9D")]
    other_hand = [Card("AD"), Card("KD"), Card("7D")]
    card = play_to_unblock(hand_to_play, other_hand)
    assert card == Card('QD')


def test_get_log_hand():
    hand_to_play = [Card('KS'), Card('4S')]
    other_hand = [Card('JS'), Card('3S'), Card('2S')]
    (long_hand, short_hand) = get_long_hand(hand_to_play, other_hand)
    assert Card('JS') in long_hand
    assert Card('4S') in short_hand



""" First seat card play."""

import random
from termcolor import termcolor

from bridgeobjects import SUITS, Card, Suit, CARD_NAMES

import bfgcardplay.src.global_variables as global_vars
from .player import Player
from .utilities import get_list_of_best_scores

MODULE_COLOUR = 'red'


class FirstSeat():

    DEFAULT_SCORE = 1
    PARTNERS_SUIT_SCORE = 5
    TENACE_SCORE = 3
    STRENGTH_SCORE = 2
    TRUMP_SCORE = 2
    FROZEN_SUIT_SCORE = 5
    LENGTH_SCORE = 2
    RUFF_AND_DISCARD_SCORE = 5
    VOID_SCORE = 1000
    DRAW_TRUMPS_SCORE = 20
    TOUCHING_HONOURS = 10
    PARTNERS_VOID = 20
    SHORTAGE_SCORE = 10

    def __init__(self, player: Player):
        self.player = player

    def update_manager(self):
        """Update the manager with played cards"""
        player = self.player
        manager = global_vars.manager
        suit = manager.suit_to_develop(player.seat)
        if suit:
            suit_strategy = manager.suit_strategy(player.seat)[suit.name]
            if suit_strategy == 'missing_ace':
                if Card('A', suit.name) not in player.total_unplayed_cards[suit.name]:
                    manager.set_suit_strategy(player.seat, suit.name, '')

            if suit_strategy == 'missing_king':
                if Card('K', suit.name) not in player.total_unplayed_cards[suit.name]:
                    manager.set_suit_strategy(player.seat, suit.name, '')

            if suit_strategy == 'missing_queen':
                if Card('Q', suit.name) not in player.total_unplayed_cards[suit.name]:
                    manager.set_suit_strategy(player.seat, suit.name, '')

            if suit_strategy == 'ruff_in_void' and player.trump_suit:
                if (player.voids[player.seat][player.trump_suit.name] or
                        player.voids[player.partner_seat][player.trump_suit.name]):
                    manager.set_suit_strategy(player.seat, player.trump_suit.name, '')

            suit = manager.suit_to_develop(player.seat)
            if suit:
                if not player.unplayed_cards[suit]:
                    manager.set_suit_to_develop(player.seat, None)

    def _deprecate_suits(self) -> list[tuple[str, int]]:
        """Assign score to suits based on void."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        for suit in SUITS:
            if not self.player.suit_cards[suit]:
                suit_scores[suit] -= self.VOID_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _partners_suit(self) -> list[tuple[str, int]]:
        """Assign score to suits based on it being suit partner led."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        partners_suit = self._get_partners_suit()
        if partners_suit and not partners_suit == self.player.trump_suit:
            suit_scores[partners_suit.name] += self.PARTNERS_SUIT_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _tenace_check(self) -> list[tuple[str, int]]:
        """Assign score to suits based on tenaces in dummy."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        dummy_tenaces = self.identify_dummy_tenaces()
        for suit in dummy_tenaces:
            if isinstance(suit, str):
                if self.player.dummy_on_left:
                    # Lead through tenaces
                    suit_scores[suit] += self.TENACE_SCORE
                elif self.player.dummy_on_right:
                    # Lead to tenaces
                    suit_scores[suit] -= self.TENACE_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _lead_through_strength(self) -> list[tuple[str, int]]:
        """Assign score to suits where you lead through or to strength."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        player = self.player
        if player.dummy_on_left:
            for suit, points in player.dummy_suit_strength.items():
                if points >= 5:
                    suit_scores[suit] += self.STRENGTH_SCORE
        elif player.dummy_on_right:
            for suit, points in player.dummy_suit_strength.items():
                if not player.trump_suit:
                    if points <= 2:
                        suit_scores[suit] += self.STRENGTH_SCORE
                else:
                    if suit != player.trump_suit.name:
                        if points <= 2:
                            suit_scores[suit] += self.STRENGTH_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _frozen_suits(self) -> list[tuple[str, int]]:
        """Assign score to suits based on frozen suits."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        frozen_suits = []
        for suit in SUITS:
            if self._frozen_suit(suit):
                frozen_suits.append(suit)
        for suit in frozen_suits:
            suit_scores[suit] -= self.FROZEN_SUIT_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _frozen_suit(self, suit: str) -> bool:
        """Return true if suit is potentially frozen."""
        cards = self.player.dummys_suit_cards[suit]
        dummy_honours = 0
        hand_honours = 0
        for card in cards:
            if card.is_honour:
                dummy_honours += 1
        if dummy_honours == 1:
            suit_cards = self.player.suit_cards[suit]
            for card in suit_cards:
                if card.is_honour:
                    hand_honours += 1
        if dummy_honours == 1 and hand_honours == 1:
            return True
        return False

    def _long_suits(self) -> list[tuple[str, int]]:
        """Assign score to suits based on length (long)."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        for suit in SUITS:
            cards = self.player.suit_cards[suit]
            number_cards = len(cards) - 4
            if number_cards > 0 and cards[0].value >= 12:
                suit_scores[suit] = number_cards * self.LENGTH_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _short_suits(self) -> list[tuple[str, int]]:
        """Assign score to suits based on length (short)."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        player = self.player
        if player.trump_suit:
            if len(player.trump_cards) > 0:
                for suit in SUITS:
                    number_cards = len(player.suit_cards[suit])
                    if 0 < number_cards < 3:
                        suit_scores[suit] = (2 - number_cards) * self.LENGTH_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _ruff_and_discard(self) -> list[tuple[str, int]]:
        """Assign score to suits based on potential for ruff and discard."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        if len(self.player.dummys_suit_cards[self.player.trump_suit.name]) > 0:
            for suit in SUITS:
                if len(self.player.dummys_suit_cards[suit]) == 0:
                    suit_scores[suit] -= self.RUFF_AND_DISCARD_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _get_partners_suit(self) -> Suit | None:
        """Return partner's suit if this is 3rd hand on first lead."""
        trick_one = self.player.board.tricks[0]
        if trick_one.leader == self.player.partner_seat:
            opening_lead = trick_one.cards[0]
            if opening_lead.suit != self.player.trump_suit:
                if opening_lead.value < 7 and self.player.suit_cards[opening_lead.suit.name]:
                    return opening_lead.suit
        return None

    def identify_dummy_tenaces(self) -> list[Suit]:
        """Return a list of suits with a tenace in dummy."""
        suits = []
        player = self.player
        for suit, tenace in player.dummy_suit_tenaces.items():
            if player.suit_cards[suit]:
                if tenace and (player.trump_suit and not suit == player.trump_suit.name):
                    tenace_index = CARD_NAMES.index(tenace.name)
                    next_card_name = CARD_NAMES[tenace_index+1]
                    if Card(next_card_name) not in player.hand_cards.list:
                        suits.append(suit)
        return suits

    def _best_suit(self, score_reasons: dict[str, int]) -> Suit:
        """Return the best suit based on score."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        player = self.player
        for reason, suits in score_reasons.items():
            for suit in suits:
                suit_scores[suit[0]] += suit[1]

        candidate_suits = get_list_of_best_scores(suit_scores)
        if not candidate_suits:
            suit = self._select_best_suit(player)
            return suit

        allowed_suits = [suit for suit in candidate_suits if player.unplayed_cards[suit]]
        return Suit(random.choice(allowed_suits))

    def _select_best_suit(self, player: Player) -> Suit:
        """Select suit for signal."""
        # TODO handle no points and equal suits
        cards = player.hand_cards.list
        strong_suits = player.get_strongest_suits(cards)
        if len(strong_suits) == 1:
            return Suit(strong_suits[0])
        # TODO get longest if multiple candidates
        long_suit_candidates = {}
        for suit_name in strong_suits:
            cards = player.suit_cards[suit_name]
            long_suit_candidates[suit_name] = len(cards)
        long_suits = get_list_of_best_scores(long_suit_candidates)
        if len(long_suits) == 1:
            return Suit(long_suits[0])
        return Suit(random.choice(long_suits))



"""Third seat card play for declarer."""

from bridgeobjects import SUITS, Card, Suit

import inspect

from ..logger import log

import bfgcardplay.src.global_variables as global_vars
from .player import Player, SuitCards
from .third_seat import ThirdSeat
from .declarer_play import card_combinations
from .utilities import trick_card_values, play_lowest_winning_card


class ThirdSeatDeclarer(ThirdSeat):
    def __init__(self, player: Player):
        super().__init__(player)

    def selected_card(self) -> Card:
        """Return the card if the third seat."""

        # Variables set in third_seat.py
        player = self.player
        trick = self.trick
        cards = self.cards

        manager = global_vars.manager

        # Void
        if not cards:
            return self._select_card_if_void()

        # Singleton
        if len(cards) == 1:
            return log(inspect.stack(), cards[0])

        # Choice of card has already been made
        if manager.card_to_play[player.seat]:
            card = manager.card_to_play[player.seat]
            manager.card_to_play[player.seat] = None
            return log(inspect.stack(), card)

        # This is a suit to develop
        suit_to_develop = manager.suit_to_develop(player.seat)
        if (suit_to_develop and
                suit_to_develop == trick.suit and
                player.suit_cards[suit_to_develop]):
            card = self._select_card_in_suit_to_develop(suit_to_develop)
            return card

        # Win trick if possible
        winning_card = self._get_winning_card()
        if winning_card:
            return winning_card

        # Cover honour with honour
        if (trick.cards[1].is_honour and
                trick.cards[1].value > trick.cards[0].value):
            for card in cards[::-1]:
                if card.value > trick.cards[1].value:
                    return log(inspect.stack(), card)

        return log(inspect.stack(), cards[-1])

    def _select_card_in_suit_to_develop(self, suit: Suit) -> Card:
        """Return the card from a suit to develop."""
        player = self.player
        trick = self.trick
        manager = global_vars.manager
        cards = player.suit_cards[suit.name]

        # Win trick if possible
        card = self._manager_win_trick(suit)
        if card:
            return card

        # Missing Ace
        card = self._missing_ace(suit)
        if card:
            return card

        # Holds all the winners
        if player.holds_all_winners_in_suit(suit, trick):
            return log(inspect.stack(), cards[0])

        # Missing King
        card = self._missing_king(suit)
        if card:
            return card

        # Missing Queen
        card = self._missing_queen(suit)
        if card:
            return card

        # Play top card if winner and suit nearly played out
        card = self._play_suit_out()
        if card:
            return card

        # Play next to top card if winner
        if (cards[0].is_honour and
                cards[0].value == cards[1].value + 1 and
                cards[1].value > trick.cards[1].value):
            return log(inspect.stack(), cards[1])

        # Play bottom of tenace if missing King
        missing_king = manager.missing_king[player.seat][suit.name]
        if player.is_tenace(cards[0], cards[1]) and missing_king:
            return log(inspect.stack(), cards[1])

        # Force out honour
        opps_unplayed_cards = player.opponents_unplayed_cards[suit.name]
        if len(cards) > 1 and len(opps_unplayed_cards) > 1:
            if (cards[1].value > opps_unplayed_cards[1].value and
                    cards[1].value > trick.cards[1].value):
                return log(inspect.stack(), cards[1])

        # Play winner if needed to make contract
        tricks_needed = player.tricks_needed - player.declarers_tricks
        if tricks_needed <= player.sure_tricks.count:
            card = play_lowest_winning_card(player, cards, trick)
            if card:
                return log(inspect.stack(), card)

        if len(opps_unplayed_cards) == 1:
            card = play_lowest_winning_card(player, cards, trick)
            if card:
                return log(inspect.stack(), card)

        unplayed_cards = player.unplayed_cards[suit.name]
        partners_unplayed_cards = player.partners_unplayed_cards[suit.name]
        if len(partners_unplayed_cards) < len(unplayed_cards):
            return log(inspect.stack(), cards[-1])

        if not player.is_winner_declarer(trick.cards[0], trick):
            for index, card in enumerate(cards[:-1]):
                if (card.value > trick.cards[1].value and
                        card. value > cards[index+1].value + 1):
                    return log(inspect.stack(), card)

        return log(inspect.stack(), cards[-1])

    def _manager_win_trick(self, suit) -> Card:
        player = self.player
        cards = self.cards
        manager = global_vars.manager
        if manager.win_trick[player.seat]:
            opps_unplayed_cards = player.opponents_unplayed_cards[suit.name]
            manager.win_trick[player.seat] = False
            for card in cards[::-1]:
                if card.value > opps_unplayed_cards[0].value:
                    return log(inspect.stack(), card)
        return None

    def _missing_ace(self, suit) -> Card:
        player = self.player
        manager = global_vars.manager
        if manager.suit_strategy(player.seat)[suit.name] == 'missing_ace':
            if not self.trick.cards[0].is_honour:
                for card in self.cards:
                    if (card.is_honour and
                            card.value > self.trick.cards[1].value):
                        return log(inspect.stack(), card)
        return None

    def _missing_king(self, suit) -> Card:
        player = self.player
        trick = self.trick
        cards = self.cards
        manager = global_vars.manager
        (ace, king, queen, jack) = SuitCards(suit.name).cards('A', 'J')
        missing_king = manager.missing_king[player.seat][suit.name]
        if missing_king:
            if ace in cards and queen not in cards and trick.cards[0] != queen:
                return log(inspect.stack(), ace)
        if manager.suit_strategy(player.seat)[suit.name] == 'missing_king':
            if trick.cards[0] == queen or trick.cards[0] == jack:
                return log(inspect.stack(), cards[-1])
        return None

    def _missing_queen(self, suit) -> Card:
        player = self.player
        manager = global_vars.manager
        (ace, king, queen, jack) = SuitCards(suit.name).cards('A', 'J')
        missing_queen = queen in player.opponents_unplayed_cards[suit.name]
        manage_strategy = manager.suit_strategy(player.seat)[suit.name]
        if (manage_strategy == 'missing_queen' and missing_queen):
            card = self._get_winning_card()
            if card:
                return card
        return None

    def _play_suit_out(self) -> Card:
        player = self.player
        if player.sure_tricks.count == 13 - player.trick_number:
            for card in self.cards[::-1]:
                if player.is_winner_declarer(card, self.trick):
                    return log(inspect.stack(), card)
        return None

    def _get_winning_card(self) -> Card | None:
        """Return a card if it can win the trick."""
        player = self.player
        trick = self.trick
        cards = self.cards

        # Trick cards
        values = trick_card_values(trick, player.trump_suit)

        # Look for tenace about a threat card and play lower card
        (higher_card, lower_card) = player.card_from_tenace_threat()
        if (lower_card and
                lower_card.value > values[0] and
                lower_card.value > values[1]):
            return log(inspect.stack(), lower_card)

        # Is this a suit to develop?
        card = self._card_from_suit_to_develop(values[0])
        if card:
            return card

        # If we hold all the winners, take the trick
        card = self._card_from_all_winners()
        if card:
            return card

        # If we probably hold all the winners, take the trick
        card = self._card_from_probable_winners()
        if card:
            return card

        # Take finesse
        card = card_combinations(player)
        if card:
            return log(inspect.stack(), card)

        card = self._get_card_over_top_of_tenace()
        if card:
            return card

        # Win trick if possible to de-block suit
        card = self._deblock_suit()
        if card:
            return card

        # Cover honour with honour
        card = self._cover_honour_with_honour()
        if card:
            return card

        # Push out honour
        if (cards[0].is_honour and
                not player.is_winner_declarer(cards[0]) and
                len(cards) > 1):
            return log(inspect.stack(), cards[1])

        # Overtake partner if appropriate
        card = self._overtake_partner()
        if card:
            return card

        # Play winner
        card = self._play_winner()
        if card:
            return card

        return None

    def _play_winner(self) -> Card:
        player = self.player
        trick = self.trick
        if not player.is_winner_declarer(trick.cards[0], trick):
            for card in self.cards[::-1]:
                if player.is_winner_declarer(card, trick):
                    return log(inspect.stack(), card)
        return None

    def _deblock_suit(self) -> Card:
        player = self.player
        trick = self.trick
        cards = self.cards
        if trick.cards[0].value < cards[0].value:
            values = trick_card_values(trick, player.trump_suit)
            is_winner = (player.is_winner_declarer(trick.cards[0], trick) and
                         values[0] > values[1])
            if (len(player.partners_unplayed_cards[trick.suit.name]) > 1 and
                    is_winner):
                return log(inspect.stack(), cards[-1])

            our_cards = player.our_unplayed_cards[trick.suit.name]
            partners_length = len(player.partners_unplayed_cards[trick.suit.name])
            my_length = len(player.unplayed_cards[trick.suit.name])
            if (player.is_winner_declarer(our_cards[0], trick) and
                    player.is_winner_declarer(our_cards[1], trick) and
                    my_length < partners_length):
                return log(inspect.stack(), cards[0])
        return None

    def _overtake_partner(self) -> Card:
        player = self.player
        trick = self.trick
        cards = self.cards
        if not player.is_winner_declarer(trick.cards[0], trick):
            values = trick_card_values(trick, player.trump_suit)
            for index, card in enumerate(cards[:-1]):
                card_value = card.value
                # trick card values already adjusted for trumps
                if card.suit == player.trump_suit:
                    card_value += 13

                if (card_value > values[0] + 1 and
                        card_value > values[1] and
                        card.value != cards[index+1].value + 1):
                    if not self._ace_is_deprecated(trick, card):
                        return log(inspect.stack(), card)
        return None

    def _cover_honour_with_honour(self) -> Card:
        trick = self.trick
        if (trick.cards[1].is_honour and
                trick.cards[1].value > trick.cards[0].value):
            for card in self.cards[::-1]:
                if (card.value > trick.cards[0].value and
                        card.value > trick.cards[1].value):
                    return log(inspect.stack(), card)

    def _get_card_over_top_of_tenace(self) -> Card:
        player = self.player
        trick = self.trick
        our_unplayed_cards = player.our_unplayed_cards[trick.suit.name]
        top_in_tenace = player.get_suit_tenaces(our_unplayed_cards)
        if top_in_tenace:
            for card in self.cards:
                if card.value < top_in_tenace.value + 1:
                    return log(inspect.stack(), card)
        return None

    def _card_from_all_winners(self) -> Card:
        player = self.player
        trick = self.trick
        cards = self.cards
        opps_unplayed_cards = player.opponents_unplayed_cards[trick.suit]
        if (player.holds_all_winners_in_suit(trick.suit, trick) or
                player.sure_tricks.count >= 13 - player.trick_number):
            if opps_unplayed_cards:
                if (trick.cards[0].value > opps_unplayed_cards[0].value and
                        trick.cards[0].value > trick.cards[1].value):
                    return log(inspect.stack(), cards[-1])
                else:
                    card = play_lowest_winning_card(player, cards, trick)
                    return log(inspect.stack(), card)
        return None

    def _card_from_probable_winners(self) -> Card:
        player = self.player
        trick = self.trick
        cards = self.cards
        opps_unplayed_cards = player.opponents_unplayed_cards[trick.suit]
        if player.holds_all_winners_in_suit(trick.suit, trick):
            if opps_unplayed_cards:
                if (trick.cards[0].value > opps_unplayed_cards[0].value and
                        trick.cards[0].value > trick.cards[1].value):
                    return log(inspect.stack(), cards[-1])
                else:
                    return log(inspect.stack(), cards[0])
        return None

    def _card_from_suit_to_develop(self, card_value) -> Card:
        player = self.player
        trick = self.trick
        cards = self.cards
        manager = global_vars.manager
        opponents_cards = []
        manage_strategy = manager.suit_strategy(player.seat)[trick.suit.name]
        if (manager.suit_to_develop(player.seat) == trick.suit or
                manage_strategy == 'missing_queen'):
            opponents_cards = player.opponents_unplayed_cards[trick.suit.name]

            card = self._card_from_tenace_in_my_hand(cards)
            if card:
                return card

            if opponents_cards:
                card = self._card_can_win(card_value, opponents_cards[0])
                if card:
                    return card
        return None

    def _card_from_tenace_in_my_hand(self, cards) -> Card:
        player = self.player
        tenace_in_my_hand = player.get_suit_tenaces(cards)
        if tenace_in_my_hand:
            for card in cards:
                if card.value < tenace_in_my_hand.value:
                    return log(inspect.stack(), card)
        return None

    def _card_can_win(self, card_value, opps_card) -> Card:
        cards = self.cards
        for card in cards[::-1]:
            if card.value > card_value + 1 and card.value > opps_card.value:
                return log(inspect.stack(), card)
        return None

    def _select_card_if_void(self) -> Card:
        """Return card if cannot follow suit."""
        player = self.player
        trick = self.trick

        # Trump if appropriate
        if player.trump_suit:
            trumps = player.trump_suit
            if not player.is_winner_declarer(trick.cards[0], trick):
                opponents_trumps = player.opponents_unplayed_cards[trumps]
                if player.trump_cards and not opponents_trumps:
                    return log(inspect.stack(), player.trump_cards[-1])
                if player.trump_cards and opponents_trumps:
                    if self._trump_partners_card():
                        card_value = trick.cards[1].value
                        if trick.cards[1].suit == player.trump_suit:
                            card_value += 13
                        trump_cards = player.unplayed_cards[player.trump_suit]
                        for card in trump_cards[::-1]:
                            if card.value + 13 > card_value:
                                return log(inspect.stack(), card)

        opponents_cards = player.opponents_unplayed_cards[trick.suit.name]

        # Find card to discard
        if player.trump_suit:
            suits = {suit_name: suit for suit_name, suit in SUITS.items()
                     if suit_name != player.trump_suit.name}
        else:
            suits = {suit_name: suit for suit_name, suit in SUITS.items()}

        # Find a loser
        suit_length: dict[str, int] = {}
        for suit_name in suits.keys():
            if player.suit_cards[suit_name]:
                suit_length[suit_name] = len(player.suit_cards[suit_name])
                our_cards = player.our_unplayed_cards[suit_name]
                opponents_cards = player.opponents_unplayed_cards[suit_name]
                if opponents_cards:
                    unplayed_cards = player.unplayed_cards[suit_name]
                    if (our_cards[0].value < opponents_cards[0].value and
                            len(unplayed_cards) > len(opponents_cards) / 2):
                        return log(inspect.stack(),
                                   player.suit_cards[suit_name][-1])
                elif (player.board.contract.denomination.is_suit and
                      not self.player.is_declarer):
                    return log(inspect.stack(),
                               player.suit_cards[suit_name][-1])

        # Only trumps left
        if not suit_length:
            return log(inspect.stack(), player.trump_cards[-1])

        # Return smallest in longest suit
        # TODO we might not want to do this
        sorted_suit_length = {
            key: value for key, value in sorted(suit_length.items(),
                                                key=lambda item: item[1],
                                                reverse=True)
            }
        long_suit = list(sorted_suit_length)[0]
        return log(inspect.stack(), player.suit_cards[long_suit][-1])



"""
    Opening leads for Card Player class.

    Returns  a card_player_components PlayedCard which includes
    name: str
    reason: str
"""

import inspect
from ..logger import log
from termcolor import cprint

from .card_player_components import SelectedCard

MODULE_COLOUR = 'blue'


def opening_lead_card(suit_cards, contract):
    """Return the opening lead from a list of cards given the contract."""
    cards = sorted(suit_cards, key=lambda x: x.order, reverse=True)
    for card in cards:
        if card.value == 9:  # T is regarded as an honour
            card.is_honour = True

    (card, selection_code) = _select_card_from_suit(cards, contract)
    return SelectedCard(card.name, selection_code)


def _are_adjacent(value_1, value_2):
    """Return True if the cards are adjacent."""
    if abs(value_1.value - value_2.value) == 1:
        return True
    return False


def _select_card_from_suit(cards, contract):
    """Return the correct card from the selected suit."""
    # Suits with 3 cards
    if len(cards) == 3:
        return _select_card_from_triplet(cards, contract)

    # Doubletons
    if len(cards) == 2:
        return _select_card_from_doubleton(cards)

    # Singletons
    if len(cards) == 1:
        return log(inspect.stack(), (cards[0], '015'))

    # Suits with 4 or more cards
    if len(cards) >= 4:
        if contract.is_nt:
            return _select_card_for_nt_contract(cards)
        return _select_card_for_suit_contract(cards)


def _select_card_for_suit_contract(cards):
    """Return the correct card for suit contract with four or more cards."""

    # NB Klinger's Guide to Better Card Play overrides Basic Bridge for 4 card suits

    # Suit headed by an A
    if cards[0].rank == 'A':
        return log(inspect.stack(), (cards[0], '013'))

    top_of_solid_sequence = solid_sequence(cards)
    if top_of_solid_sequence:
        return log(inspect.stack(), (top_of_solid_sequence, '020'))

    top_of_near_sequence = near_sequence(cards)
    if top_of_near_sequence:
        return log(inspect.stack(), (top_of_near_sequence, '021'))

    top_of_internal_sequence = internal_sequence(cards)
    if top_of_internal_sequence:
        return log(inspect.stack(), (top_of_internal_sequence, '022'))

    top_of_touching_honours = touching_honours(cards)
    if top_of_touching_honours:
        return log(inspect.stack(), (top_of_touching_honours, '023'))

    # With 4 rags, lead second highest
    if len(cards) >= 4:
        if not cards[0].is_honour:
            if len(cards) > 4:
                return log(inspect.stack(), (cards[3], '004'))
            return log(inspect.stack(), (cards[1], '014'))

    # With an honour, lead fourth highest
    return log(inspect.stack(), (cards[3], '004'))


def _select_card_for_nt_contract(cards):
    """Return the correct card in a NT contract with four or more cards."""

    top_of_solid_sequence = solid_sequence(cards)
    if top_of_solid_sequence:
        return log(inspect.stack(), (top_of_solid_sequence, '020'))

    top_of_near_sequence = near_sequence(cards)
    if top_of_near_sequence:
        return log(inspect.stack(), (top_of_near_sequence, '021'))

    # if top_of_near_sequence:
    #     return log(inspect.stack(), (top_of_near_sequence, '021'))

    top_of_internal_sequence = internal_sequence(cards)
    if top_of_internal_sequence:
        return log(inspect.stack(), (top_of_internal_sequence, '022'))

    # With 4 rags, lead second highest
    if len(cards) >= 4:
        if not cards[0].is_honour:
            if len(cards) > 4:
                return log(inspect.stack(), (cards[3], '004'))
            return log(inspect.stack(), (cards[1], '014'))

    # Lead fourth highest
    return log(inspect.stack(), (cards[3], '004'))


def _select_card_from_triplet(cards, contract):
    """Return the correct card from a three card suit."""

    # Suit headed by A and no K in a NT contract
    if cards[0].rank == 'A' and cards[1].rank != 'K':
        if contract.is_nt:
            return log(inspect.stack(), (cards[2], '007'))
        return log(inspect.stack(), (cards[0], '013'))

    if cards[0].rank == 'A' and cards[1].rank == 'K':
        return log(inspect.stack(), (cards[0], '013'))

    # Suit headed by an honour and adjacent card
    if cards[0].is_honour and _are_adjacent(cards[0], cards[1]):
        return log(inspect.stack(), (cards[0], '003'))

    # Suit headed by an honour: top of two touching cards
    if (cards[1].is_honour and _are_adjacent(cards[1], cards[2])):
        return log(inspect.stack(), (cards[1], '008'))

    # Suit headed by two honours
    if (cards[0].is_honour and cards[1].is_honour):
        return log(inspect.stack(), (cards[2], '010'))

    # Suit headed by a single honour
    if (cards[0].is_honour and not cards[1].is_honour):
        return log(inspect.stack(), (cards[2], '007'))

    # Return middle card
    return log(inspect.stack(), (cards[1], '005'))


def _select_card_from_doubleton(cards):
    """Return the correct card from a two card suit."""
    if cards[0].rank == 'A' and cards[1].rank == 'K':
        return log(inspect.stack(), (cards[1], '002'))
    return log(inspect.stack(), (cards[0], '001'))


def solid_sequence(cards):
    """Returns the top of the solid sequence if there is one, or None."""
    if len(cards) < 3:
        return None
    for index in range(len(cards)-2):
        if (cards[index].value == cards[index+1].value + 1 and
                cards[index+1].value == cards[index+2].value + 1 and
                cards[index].value >= 9):
            return cards[index]
    return None


def near_sequence(cards):
    """Returns the top of the near sequence if there is one, or None."""
    if len(cards) < 3:
        return None
    for index in range(len(cards)-2):
        if (cards[index].value == cards[index+1].value + 1 and
                cards[index+1].value == cards[index+2].value + 2 and
                cards[index].value >= 9):
            return cards[index]
    return None


def internal_sequence(cards):
    """Returns the top of the internal sequence if there is one, or None."""
    if len(cards) < 3:
        return None
    for index in range(len(cards) - 2):
        card = _find_internal_sequence(cards, index)
        if card:
            return card
    return None


def _find_internal_sequence(cards, index):
    for offset in range(2, 5):  # Check Q, J and T
        if (cards[index].value == cards[index+1].value + offset and
                cards[index+1].value == cards[index+2].value + 1 and
                cards[index+1].is_honour):
            return cards[index+1]
    return None


def touching_honours(cards):
    """Returns the top of the touching honours if there is one, or None."""
    if len(cards) < 3:
        return None
    for index in range(len(cards)-1):
        if (cards[index].value == cards[index+1].value + 1 and
                cards[index].value >= 12):
            return cards[index]
    return None



"""The manager supervises play and holds persistent data.
It is created in global.py."""

from bridgeobjects import SEATS, SUITS, RANKS, Suit

MODULE_COLOUR = 'magenta'

DECLARER_STRATEGY = {
    0: 'Draw trumps and establish long suit',
}

ranks = [rank for rank in RANKS]
ranks.reverse()
RANKS = ranks[:-1]


class Manager():
    def __init__(self):
        # TODO remove print statements
        # print('')
        # print('*'*50)
        # print('')
        self.initialise()

    def initialise(self):
        """Initialise variables at start of board."""
        # print('manager initialised')
        self._board = None
        self.threats = None
        # self.winning_suit = None
        self.long_player = None
        self.draw_trumps = False
        self.missing_king = {seat: {suit_name: False
                                    for suit_name in SUITS} for seat in SEATS}
        self.working_suit = {seat: None for seat in SEATS}
        self._suit_to_develop = {seat: None for seat in SEATS}
        self._suit_strategy = {seat: {suit: ''
                                      for suit in SUITS} for seat in SEATS}
        self.card_to_play = {seat: [] for seat in SEATS}
        self.voids = {seat: {suit_name: False
                             for suit_name in SUITS} for seat in SEATS}
        self.win_trick = {seat: False for seat in SEATS}
        # Defenders
        self.signal_card = {seat: None for seat in SEATS}
        self._like_dislike = {seat: {suit: 0
                                     for suit in SUITS} for seat in SEATS}
        self._even_odd = {seat: {suit: 0 for suit in SUITS} for seat in SEATS}

    def __repr__(self):
        return f'Manager {self.suit_to_develop("N")} {self.suit_strategy("N")}'

    @property
    def board(self):
        """Return the current Board."""
        return self._board

    @board.setter
    def board(self, value):
        """Set the value of the current board"""
        self._board = value

    def suit_to_develop(self, seat: str) -> Suit:
        """Return the suit to develop for a seat."""
        return self._suit_to_develop[seat]

    def set_suit_to_develop(self, seat: str, suit: Suit):
        """Set the suit to develop for a seat."""
        partners_seat = self._partners_seat(seat)
        self._suit_to_develop[seat] = suit
        self._suit_to_develop[partners_seat] = suit

    def suit_strategy(self, seat: str) -> str:
        """Return the suit to develop for a seat."""
        return self._suit_strategy[seat]

    def set_suit_strategy(self, seat: str, suit: str, strategy: str):
        """Set the suit to develop for a seat."""
        partners_seat = self._partners_seat(seat)
        self._suit_strategy[seat][suit] = strategy
        self._suit_strategy[partners_seat][suit] = strategy

    def set_missing_king(self, seat: str, suit: str, missing_king: bool):
        """Set missing king for the suit for a seat."""
        partners_seat = self._partners_seat(seat)
        self.missing_king[seat][suit] = missing_king
        self.missing_king[partners_seat][suit] = missing_king

    def like_dislike(self, seat: str, suit: str) -> int:
        """Return 1 if the player likes the suit and -1 if not."""
        return self._like_dislike[seat][suit]

    def set_like_dislike(self, seat: str, suit: str, value: bool):
        """Set the like_dislike for a seat and suit."""
        self._like_dislike[seat][suit] = value

    def _even_odd(self, seat: str, suit: str) -> bool | None:
        """Return 1 if the player has an even number the suit and -1 if not."""
        return self._even_odd[seat][suit]

    def set_even_odd(self, seat: str, suit: str, value: int):
        """Set even/odd (0/1) for a seat and suit."""
        self._even_odd[seat][suit] = value

    @staticmethod
    def _partners_seat(seat: str) -> str:
        """Return partners's seat name"""
        seat_index = SEATS.index(seat)
        partners_index = (seat_index + 2) % 4
        partners_seat = SEATS[partners_index]
        return partners_seat



# first_seat_declarer.py

""" First seat card play for declarer."""

from termcolor import cprint

import inspect
from ..logger import log

from bridgeobjects import SUITS, Card, Suit

import bfgcardplay.src.global_variables as global_vars
from .player import Player, Cards
from .first_seat import FirstSeat

MODULE_COLOUR = 'cyan'


ALL_WINNERS = 5


class FirstSeatDeclarer(FirstSeat):
    def __init__(self, player: Player):
        super().__init__(player)

    def selected_card(self) -> Card:
        """Return the selected card for the first seat."""
        player = self.player
        manager = global_vars.manager
        if player.trump_suit:
            suit = self._select_suit_for_suit_contract(manager)
        else:
            suit = self._select_suit_for_nt_contract(manager)
        if manager.card_to_play[player.seat]:
            card = manager.card_to_play[player.seat][0]
            manager.card_to_play[player.seat].pop(0)
            if manager.suit_to_develop[player.seat] and manager.card_to_play[player.seat]:
                manager.card_to_play[player.seat] = player.suit_cards[manager.suit_to_develop][-1]
            log(inspect.stack(), f'{card}')
            return card
        if not player.suit_cards[suit.name]:
            raise ValueError(f'No cards for {suit}')
        if manager.winning_suit:
            if player.suit_cards[manager.winning_suit.name]:
                return self._select_card_from_manager_winning_suit(manager)
        card = self._select_lead_card(manager, suit)
        log(inspect.stack(), f'{card}')
        return card

    def _select_suit_for_suit_contract(self, manager) -> Suit:
        """Return the trick lead suit for the declarer in  a suit contract."""
        player = self.player
        manager = global_vars.manager
        score_reasons = {}

        # Draw trumps?
        if player.opponents_trumps and self._can_draw_trumps():
            return player.trump_suit
        else:
            score_reasons['trumps'] = self._deprecate_trumps()
            winners = self._all_winners_score()
            total_winners = 0
            for suit_count in winners:
                total_winners += suit_count[1]
            if total_winners <= 13 - player.trick_number:
                score_reasons['shortage'] = self._shortage_score(player.controls)
                suit_name = manager.suit_to_develop[player.seat].name
                if suit_name and player.suit_cards[suit_name]:
                    return manager.suit_to_develop[player.seat]
                if not player.suit_cards[suit_name]:
                    manager.suit_to_develop[player.seat] = None

        winning_suit = self._play_winning_suit(manager)
        if winning_suit:
            return winning_suit

        if player.trump_suit and not player.opponents_trumps:
            long_suits = player.partnership_long_suits()
            for suit in long_suits:
                if player.can_lead_toward_tenace(suit):
                    return Suit(suit)

        return player.longest_suit

    def _select_suit_for_nt_contract(self, manager) -> Suit:
        """Return the trick lead suit for the declarer in  a suit contract."""
        player = self.player

        winners = player.get_winners()
        target = player.declarers_target
        our_tricks = player.declarers_tricks
        # print(f'{target=}')
        # print(f'{winners=}')
        # print(f'{our_tricks=}')
        # cprint(f'{controls = }', 'red')

        if target - (winners + our_tricks) > 0:
            if len(player.board.tricks) == 2:
                suit_to_develop = self._suits_to_develop(manager, player.controls)
                if suit_to_develop:
                    return suit_to_develop

        winning_suit = self._play_winning_suit(manager)
        if winning_suit:
            return winning_suit

        # long_suits = player.partnership_long_suits()
        # for suit in long_suits:
        #     if player.can_lead_toward_tenace(suit):
        #         return Suit(suit)

        score_reasons = {}

        # Deprecate voids
        score_reasons['void'] = self._deprecate_suits()

        # All cards are winners
        score_reasons['all_winners'] = self._all_winners_score()

        # Avoid frozen suits
        score_reasons['frozen'] = self._frozen_suits()

        # Long suits
        score_reasons['long'] = self._long_suits()

        # Controls
        score_reasons['controls'] = self._long_suits()

        # Suit strength
        score_reasons['strength'] = self._suit_strength()

        # Select best suit
        best_suit = self._best_suit(score_reasons)
        # print(colored(f'{best_suit=} {score_reasons=}', 'red'))
        return best_suit

    def _suits_to_develop(self, manager: object, controls: list[str]) -> list[str]:
        """Return a list of suits to develop."""
        player = self.player
        suits_to_develop = []
        suit_to_develop = None
        for suit_name in SUITS:
            if not controls[suit_name]:
                suits_to_develop.append(suit_name)
        long_suits_to_develop = []

        lead_to_this_hand = False
        for suit_name in suits_to_develop:
            declarers_suit_cards = player.declarers_suit_cards[suit_name]
            dummys_suit_cards = player.dummys_suit_cards[suit_name]
            total_cards = len(declarers_suit_cards) + len(dummys_suit_cards)
            if (total_cards >= 6 and
                    (len(declarers_suit_cards) >= 4 or len(dummys_suit_cards) >= 4)):
                long_suits_to_develop.append(suit_name)
                my_touching_honours = player.hand.touching_honours()
                partner_touching_honours = player.partners_hand.touching_honours()
                for suit_name in suits_to_develop:

                    suit_cards = player.partners_suit_cards[suit_name]
                    if player.get_suit_tenaces(suit_cards):
                        suit_to_develop = suit_name

                    suit_cards = player.suit_cards[suit_name]
                    if player.get_suit_tenaces(suit_cards):
                        suit_to_develop = suit_name
                        lead_to_this_hand = True
                    if partner_touching_honours[suit_name]:
                        suit_to_develop = suit_name
                    if my_touching_honours[suit_name]:
                        suit_to_develop = suit_name
                        lead_to_this_hand = True

        entry = None
        if lead_to_this_hand:
            partners_entries = player.get_entries(player.partners_hand)
            for suit_name in SUITS:
                if partners_entries[suit_name] and player.suit_cards[suit_name]:
                    entry = partners_entries[suit_name][0]
                    break
        if entry:
            if suit_to_develop:
                manager.suit_to_develop[player.declarer] = suit_to_develop
                manager.suit_to_develop[player.dummy] = suit_to_develop
                manager.card_to_play[player.partner_seat].append(entry)
                card = player.partners_suit_cards[suit_to_develop][-1]
                manager.card_to_play[player.partner_seat].append(card)
                manager.card_to_play[player.seat].append(player.suit_cards[entry.suit.name][-1])
                return Suit(suit_to_develop)
        return None

    def _select_lead_card(self, manager, suit: Suit) -> Card:
        """Return the selected lead card for declarer."""
        player = self.player
        cards = player.suit_cards[suit.name]
        # Top of touching honours
        for index, card in enumerate(cards[:-1]):
            if card.is_honour and card.value == cards[index+1].value + 1:
                log(inspect.stack(), f'{card}')
                return card

        # Top of doubleton
        if len(cards) == 2:
            log(inspect.stack(), f'{cards[0]}')
            return cards[0]

        # Return bottom card
        log(inspect.stack(), f'{cards[-1]}')
        return cards[-1]

    def _select_card_from_manager_winning_suit(self, manager: object) -> Card:
        """Return the selected lead card from manager winning suit."""
        player = self.player
        suit = manager.winning_suit
        long_player = player.long_seat(suit.name)
        manager.long_player = long_player
        cards = player.suit_cards[suit.name]
        if long_player != player.seat:
            # Short hand leads top card
            log(inspect.stack(), f'{cards[0]}')
            return cards[0]
        # Long hand leads bottom card if short hand has cards
        if player.partners_suit_cards[suit.name]:
            log(inspect.stack(), f'{cards[-1]}')
            return cards[-1]
        else:
            log(inspect.stack(), f'{cards[0]}')
            return cards[0]

    def _play_winning_suit(self, manager) -> Suit:
        """Return the winning suit or None."""
        player = self.player
        # Reset Manager winning suit when no cards left
        if manager.winning_suit:
            if player.suit_cards[manager.winning_suit.name]:
                return manager.winning_suit
            else:
                manager.winning_suit = None

        winning_suit = self._get_winning_suit(player)
        if winning_suit:
            manager.winning_suit = winning_suit
            return winning_suit
        return None

    def _get_winning_suit(self, player: Player) -> Suit:
        """Return the selected winning suit."""
        winning_suits = self._get_winning_suits(player)
        max_length = 0
        long_suit = None
        for suit in winning_suits:
            suit_length = len(player.suit_cards[suit.name])
            if suit_length > max_length:
                max_length = suit_length
                long_suit = suit
        return long_suit

    def _get_winning_suits(self, player: Player) -> list[Suit]:
        """Return a list of winning suits."""
        winning_suits = []
        for suit in SUITS.values():
            if suit != player.trump_suit:
                if player.holds_all_winners_in_suit(suit):
                    winning_suits.append(suit)
        return winning_suits

    def _can_draw_trumps(self) -> bool:
        """Return True if declarer should draw trumps."""
        player = self.player
        if not player.trump_cards:
            return False

        declarers_trumps = Cards(player.declarers_hand.cards).by_suit[player.trump_suit]
        dummys_trumps = Cards(player.dummys_hand.cards).by_suit[player.trump_suit]
        if (len(declarers_trumps) + len(dummys_trumps) <= 7 and
                max(len(declarers_trumps), len(dummys_trumps)) == 4):
            return False

        declarers_trumps = player.unplayed_cards_by_suit(player.trump_suit, player.declarer)
        dummys_trumps = player.unplayed_cards_by_suit(player.trump_suit, player.dummy)
        if len(declarers_trumps) + len(dummys_trumps) >= 9:  # TODO too simplistic - determine losers first
            return True

        declarer_longer_trumps = len(declarers_trumps) > len(dummys_trumps)
        if declarer_longer_trumps:
            short_hand = player.board.hands[player.dummy]
        else:
            short_hand = player.board.hands[player.declarer]
        if short_hand.shape[3] <= 2:
            # Are there sufficient rumps in short hand?
            opponents_trumps = player.opponents_unplayed_cards[player.trump_suit.name]
            if len(short_hand.cards_by_suit[player.trump_suit.name]) > len(opponents_trumps):
                return True
            # can declarer produce a void in time?
            suit = short_hand.shortest_suit
            # if declarer_longer_trumps:
            #     short_suit_cards = player.unplayed_cards_by_suit(suit, player.dummy)
            # else:
            #     short_suit_cards = player.unplayed_cards_by_suit(suit, player.declarer)

            declarers_cards = player.unplayed_cards_by_suit(suit, player.declarer)
            dummys_cards = player.unplayed_cards_by_suit(suit, player.dummy)
            suit_cards = declarers_cards + dummys_cards

            if (Card('A', suit.name) not in suit_cards or Card('K', suit.name) in suit_cards):
                return False
        return True

    def _long_suits(self) -> list[tuple[str, int]]:
        """Assign score to suits based on length (long)."""
        player = self.player
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        length_score = self.LENGTH_SCORE

        for suit in SUITS:
            number_cards = len(player.our_cards[suit]) - len(player.opponents_cards[suit])
            if number_cards > 0:
                if player.trump_suit:
                    if suit == player.trump_suit.name:
                        length_score = 0
                suit_scores[suit] = number_cards * length_score
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    # def _frozen_suits(self) -> list[tuple[str, int]]:
    #     """Assign score to suits based on frozen suits."""
    #     player = self.player
    #     suit_scores = {suit_name: 0 for suit_name, suit in SUITS.items()}
    #     frozen_suits = []
    #     dummys_cards = player.board.hands[player.dummy].cards
    #     for suit, cards in player.get_cards_by_suit(dummys_cards).items():
    #         dummy_honours = 0
    #         hand_honours = 0
    #         for card in cards:
    #             if card.is_honour:
    #                 dummy_honours += 1
    #         if dummy_honours == 1:
    #             declarers_cards = player.board.hands[player.declarer].cards
    #             for card in declarers_cards:
    #                 if card.is_honour:
    #                     hand_honours += 1
    #         if dummy_honours ==1 and hand_honours == 1:
    #             frozen_suits.append(suit)
    #     for suit in frozen_suits:
    #         suit_scores[suit] -= self.FROZEN_SUIT_SCORE
    #     return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _all_winners_score(self) -> list[tuple[str, int]]:
        """Assign score to a suit if all the cards are winners."""
        player = self.player
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        for suit_name, suit in SUITS.items():
            if suit != player.trump_suit:
                if player.holds_all_winners_in_suit(suit):
                    suit_scores[suit_name] = ALL_WINNERS
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _deprecate_trumps(self) -> list[tuple[str, int]]:
        """Assign score to a suit if opponents have no trumps."""
        suit_scores = {suit_name: 0 for suit_name, suit in SUITS.items()}
        suit_scores[self.player.trump_suit.name] = -1 * self.TRUMP_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _suit_strength(self) -> list[tuple[str, int]]:
        """Assign score to a suit by HCP."""
        suit_scores = {suit_name: 0 for suit_name, suit in SUITS.items()}
        for suit_name in SUITS:
            cards = self.player.our_cards[suit_name]
            for card in cards:
                if card.value >= 10:
                    suit_scores[suit_name] += card.value - 9
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _shortage_score(self, controls: dict[str, int]) -> list[tuple[str, int]]:
        """Assign score to a suit if we can create a shortage."""
        player = self.player
        manager = global_vars.manager
        suit_scores = {suit_name: 0 for suit_name, suit in SUITS.items()}
        for suit_name, suit in SUITS.items():
            if suit != player.trump_suit:
                if controls[suit_name] - len(player.partners_suit_cards[suit_name]) <= 1:
                    suit_scores[suit_name] += self.SHORTAGE_SCORE
                    manager.suit_to_develop[player.declarer] = Suit(suit_name)
                    manager.suit_to_develop[player.dummy] = Suit(suit_name)
        return [(suit_name, score) for suit_name, score in suit_scores.items()]



"""Methods to support declarer play."""
from termcolor import cprint

import inspect
from ..logger import log

from bridgeobjects import Card, Suit, CARD_VALUES
from bfgdealer import Trick

import bfgcardplay.src.global_variables as global_vars
from .player import Player
from .dashboard import SuitCards

MODULE_COLOUR = 'magenta'


def card_combinations(player: Player, suit: str = '') -> Card | None:
    """Return a card from a defined combination, or None."""
    trick = player.board.tricks[-1]
    if not suit:
        suit = trick.suit.name
    opponents_cards = player.opponents_unplayed_cards[suit]
    our_cards = player.our_unplayed_cards[suit]
    suit_cards = SuitCards(suit).cards('A', 'J')
    (ace, king, queen, jack) = suit_cards

    if (ace in our_cards and
            (queen in our_cards or jack in our_cards) and
            king in opponents_cards):
        return missing_king_new(player, suit)

    if (ace in our_cards and
            # (king in our_cards or jack in our_cards) and
            queen in opponents_cards):
        return missing_queen_new(player, suit)
    return None


def lead_to_find_missing_ace(player: Player,
                             suit: Suit,
                             trick: Trick | None = None) -> Card | None:
    """Return the appropriate card if missing the ace."""
    manager = global_vars.manager
    my_suit_cards = player.suit_cards[suit]
    partners_suit_cards = player.partners_suit_cards[suit]
    max_length = max(len(my_suit_cards), len(partners_suit_cards))
    if max_length >= 4:
        if (Card('K', suit.name) in player.our_unplayed_cards[suit.name] and
                Card('Q', suit.name) in player.our_unplayed_cards[suit.name] and
                Card('J', suit.name) in player.our_unplayed_cards[suit.name] and
                Card('T', suit.name) in player.our_unplayed_cards[suit.name]):
            for card in my_suit_cards:
                if card.value >= CARD_VALUES['T']:
                    return log(inspect.stack(), card)

    partners_entries = player.get_entries_in_other_suits(player.partners_hand, suit)
    if ((Card('K', suit.name) in my_suit_cards and
            Card('Q', suit.name) in my_suit_cards and partners_entries) or
            (Card('Q', suit.name) in my_suit_cards and
             Card('J', suit.name) in my_suit_cards and partners_entries)):
        card = _make_entry_to_partners_hand(player, manager, partners_entries)
        if card:
            return log(inspect.stack(), card)
    return my_suit_cards[-1]


def _make_entry_to_partners_hand(player, manager, partners_entries):
    """Return a card to enable partner to return suit_to_develop."""
    for entry in partners_entries:
        entry_suit = entry.suit.name
        my_entry_cards = player.suit_cards[entry_suit]
        if my_entry_cards:
            manager.win_trick[player.partner_seat] = True
            return my_entry_cards[-1]
    return None


def missing_king(player: Player, suit: Suit) -> Card | None:
    """Return the appropriate card if missing the king."""
    manager = global_vars.manager
    my_suit_cards = player.suit_cards[suit]
    partners_suit_cards = player.partners_suit_cards[suit]
    my_entries = player.get_entries_in_other_suits(player.hand, suit)
    partners_entries = player.get_entries_in_other_suits(player.partners_hand, suit)

    tenace_in_my_hand = player.get_suit_tenaces(my_suit_cards)
    right_hand_opponent_is_void = player.voids[player.right_hand_seat][suit.name]
    if tenace_in_my_hand and right_hand_opponent_is_void:
        for card in my_suit_cards:
            if card.value < CARD_VALUES['K']:
                return log(inspect.stack(), card)

    if tenace_in_my_hand and len(partners_entries) >= 1:
        card = _make_entry_to_partners_hand(player, manager, partners_entries)
        if card:
            return log(inspect.stack(), card)

    tenace_in_partners_hand = player.get_suit_tenaces(partners_suit_cards)
    queen = Card('Q', suit.name)
    jack = Card('J', suit.name)
    if tenace_in_partners_hand and len(my_entries) >= 1:
        return log(inspect.stack(), my_suit_cards[-1])
    elif queen in my_suit_cards and jack in my_suit_cards:
        return log(inspect.stack(), queen)
    elif queen in partners_suit_cards and jack in partners_suit_cards:
        return log(inspect.stack(), my_suit_cards[-1])

    for card in my_suit_cards[::-1]:
        if player.is_winner_declarer(card):
            return log(inspect.stack(), card)
        else:
            return log(inspect.stack(), my_suit_cards[0])

    cprint(f'missing_king return None {suit}', MODULE_COLOUR)
    return None


def missing_king_new(player: Player, suit: str) -> Card | None:
    """Return the appropriate card if missing the king."""
    manager = global_vars.manager
    manager.set_missing_king(player.seat, suit, True)

    our_card_count = len(player.our_cards[suit])
    if our_card_count >= 11:
        return _missing_king_with_eleven_cards(player, suit)
    if our_card_count == 10:
        return _missing_king_with_ten_cards(player, suit)
    elif our_card_count == 9:
        return _missing_king_with_nine_cards(player, suit)
    elif our_card_count == 8:
        return _missing_king_with_eight_cards(player, suit)
    elif 5 <= our_card_count <= 8:
        return _missing_king_with_five_to_eight_cards(player, suit)

    return None


def _missing_king_with_eleven_cards(player, suit):
    my_cards = player.unplayed_cards[suit]
    (ace, king, queen, jack, ten) = SuitCards(suit).cards('A', 'T')
    if ace in my_cards:
        return log(inspect.stack(), ace)
    return log(inspect.stack(), my_cards[-1])


def _missing_king_with_ten_cards(player, suit):
    my_cards = player.unplayed_cards[suit]
    partners_cards = player.partners_unplayed_cards[suit]
    (ace, king, queen, jack, ten) = SuitCards(suit).cards('A', 'T')

    # Partner holds A and Q
    if ace in partners_cards and queen in partners_cards:
        return log(inspect.stack(), my_cards[-1])

    # A in this hand and partner holds the Q
    if ace in my_cards and queen in partners_cards:
        return log(inspect.stack(), ace)

    # QJ in this hand and partner holds the A
    if (ace in partners_cards and queen in my_cards and jack in my_cards):
        return log(inspect.stack(), queen)

    # Q in this hand and partner holds the A
    if ace in partners_cards and queen in my_cards:
        return log(inspect.stack(), my_cards[-1])

    # A and Q in this hand
    if ace in my_cards and queen in my_cards:
        return None
    return None


def _missing_king_with_nine_cards(player, suit):
    my_cards = player.unplayed_cards[suit]
    partners_cards = player.partners_unplayed_cards[suit]
    (ace, king, queen, jack, ten) = SuitCards(suit).cards('A', 'T')
    # Partner holds A and Q and this hand holds JT
    if (ace in partners_cards and queen in partners_cards and
            jack in my_cards and ten in my_cards):
        return log(inspect.stack(), jack)

    # Partner holds JT and this hand holds AQ
    if (ace in my_cards and queen in my_cards and
            jack in partners_cards and ten in partners_cards):
        return None

    # Partner holds A and Q
    if ace in partners_cards and queen in partners_cards:
        return log(inspect.stack(), my_cards[-1])

    # A in this hand
    if ace in my_cards:
        return log(inspect.stack(), ace)

    # Partner holds A
    if ace in partners_cards:
        return log(inspect.stack(), my_cards[-1])
    return None


def _missing_king_with_eight_cards(player, suit):
    my_cards = player.unplayed_cards[suit]
    partners_cards = player.partners_unplayed_cards[suit]
    (ace, king, queen, jack, ten) = SuitCards(suit).cards('A', 'T')
    # Partner holds A and Q
    if ace in partners_cards and queen in partners_cards:
        return log(inspect.stack(), my_cards[-1])
    return None


def _missing_king_with_five_to_eight_cards(player, suit):
    my_cards = player.unplayed_cards[suit]
    partners_cards = player.partners_unplayed_cards[suit]
    # We hold AQJT9
    our_cards = my_cards + partners_cards
    (ace, king, queen, jack, ten) = SuitCards(suit).cards('A', 'T')
    nine = Card('9', suit)
    if (ace in our_cards and queen in our_cards and
            jack in our_cards and ten in our_cards and nine in our_cards):
        if len(my_cards) > 1:
            return log(inspect.stack(), my_cards[1])
        else:
            return log(inspect.stack(), my_cards[-1])
    return None


def missing_queen_new(player: Player, suit: str) -> Card | None:
    """Return the appropriate card if missing the queen."""
    trick = player.board.tricks[-1]
    (ace, king, queen, jack, ten) = SuitCards(suit).cards('A', 'T')
    my_cards = player.unplayed_cards[suit]
    partners_cards = player.partners_unplayed_cards[suit]

    # hand holds AJ and partner K
    if ace in my_cards and jack in my_cards and king in partners_cards:
        if len(my_cards) > 2:
            return log(inspect.stack(), my_cards[1])

    # hand holds K and partner AJ
    elif king in my_cards and ace in partners_cards and jack in partners_cards:
        return log(inspect.stack(), king)

    # hand holds KJ and partner A
    elif king in my_cards and jack in my_cards and ace in partners_cards:
        if len(my_cards) > 2:
            return my_cards[-1]
        return log(inspect.stack(), my_cards[-1])

    # hand holds AJ and K not played by partner
    elif ace in my_cards and jack in my_cards and trick.cards and trick.cards[0] != king:
        return log(inspect.stack(), jack)

    # hand holds A and K not played by partner
    elif ace in my_cards and trick.cards and trick.cards[0] != king:
        return log(inspect.stack(), ace)
    return None


def missing_queen(player: Player,
                  suit: Suit,
                  trick: Trick | None = None) -> Card | None:
    """Return the appropriate card if missing the queen."""
    manager = global_vars.manager
    my_suit_cards = player.suit_cards[suit]
    partners_suit_cards = player.partners_suit_cards[suit]
    (ace, king, queen, jack, ten) = SuitCards(suit.name).cards('A', 'T')

    if not partners_suit_cards:
        return log(inspect.stack(), my_suit_cards[0])

    if len(player.our_cards[suit.name]) >= 9:
        if trick:
            trick_cards = trick.cards
        else:
            trick_cards = []
        if ace in my_suit_cards and king not in trick_cards:
            return log(inspect.stack(), my_suit_cards[0])
        if king in my_suit_cards and ace not in trick_cards:
            return log(inspect.stack(), my_suit_cards[0])
        return log(inspect.stack(), my_suit_cards[-1])

    my_entries = player.get_entries_in_other_suits(player.hand, suit)
    partners_entries = player.get_entries_in_other_suits(player.partners_hand, suit)

    tenace_in_my_hand = player.get_suit_tenaces(my_suit_cards)
    right_hand_opponent_is_void = player.voids[player.right_hand_seat][suit.name]
    if tenace_in_my_hand and right_hand_opponent_is_void:
        for card in my_suit_cards:
            if card.value < CARD_VALUES['Q']:
                return log(inspect.stack(), card)

    if tenace_in_my_hand and len(partners_entries) >= 1:
        card = _make_entry_to_partners_hand(player, manager, partners_entries)
        if card:
            return log(inspect.stack(), card)

    if ace in partners_suit_cards:
        return log(inspect.stack(), my_suit_cards[-1])

    if not tenace_in_my_hand:
        tenace_in_partners_hand = player.get_suit_tenaces(partners_suit_cards)
        if tenace_in_partners_hand and len(my_entries) >= 1:
            if my_suit_cards[0] == ace or my_suit_cards[0] == king:
                return log(inspect.stack(), my_suit_cards[0])
            else:
                return log(inspect.stack(), my_suit_cards[-1])

    for card in my_suit_cards[::-1]:
        if player.is_winner_declarer(card):
            return log(inspect.stack(), card)
        else:
            partners_cards = player.partners_unplayed_cards[suit.name]
            if not partners_cards:
                return None
            manager.card_to_play[player.partner_seat] = partners_cards[-1]
            return log(inspect.stack(), my_suit_cards[0])
    cprint(f'missing_queen return None {suit}', MODULE_COLOUR)
    return None
